import logging
import select
import socket
import struct
from typing import Iterable, Dict, NamedTuple, Type, List, Callable, Optional
from urllib.parse import ParseResult, urlparse, parse_qs

from xrpc.net import RPCPacket

Origin = str

PACKET_PACKER = struct.Struct('!I')


class Packet(NamedTuple):
    addr: Origin
    data: bytes

    def pack(self) -> bytes:
        body = self.data

        len_bts = PACKET_PACKER.pack(len(body))

        return len_bts + body

    @classmethod
    def unpack(cls, addr: Origin, buffer: bytes):
        buffer = memoryview(buffer)

        if len(buffer) < 4:
            raise ValueError('Could not parse size')
        size, = PACKET_PACKER.unpack(buffer[:4])
        if len(buffer) < 4 + size:
            raise ValueError('Size is not enough')

        y = Packet(addr, buffer[4:size + 4].tobytes())

        return y


class Transport:
    def __init__(self, url):
        self.url: ParseResult = urlparse(url)
        self._fd = None

    @property
    def fd(self):
        assert self._fd, '_fd must be set'
        return self._fd

    @fd.setter
    def fd(self, value):
        self._fd = value

    def send(self, packet: Packet):
        raise NotImplementedError()

    def read(self) -> Iterable[Packet]:
        raise NotImplementedError()

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def from_url(cls, url) -> 'Transport':
        parsed: ParseResult = urlparse(url)

        if parsed.scheme in TRANSPORT_MAP:
            return TRANSPORT_MAP[parsed.scheme](url)
        else:
            raise ValueError(f'Could not find transport for scheme `{parsed.scheme}`')


def recvfrom_helper(fd, buffer_size=2 ** 16, logger_name='net.trace.raw'):
    try:
        while True:

            buffer, addr = fd.recvfrom(buffer_size)

            if len(buffer) == 0:
                raise ConnectionAbortedError('Zero bytes received')

            try:
                yield Packet.unpack(addr, buffer)
            except ValueError:
                logging.getLogger(f'{logger_name}.e').error('[%d] %s %s', len(buffer), addr, buffer)
                break
    except BlockingIOError:
        return


class UDPTransport(Transport):
    # we need this on both sides, really
    # Sender and Receiver
    def __init__(self, url, buffer=2 ** 16):
        super().__init__(url)

        assert isinstance(buffer, int), buffer

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP

        try:
            sock.bind((self.url.hostname if self.url.hostname else '0.0.0.0', self.url.port if self.url.port else 0))
        except OSError as e:
            raise OSError(f'{url}')
        sock.settimeout(0)

        self.fd: socket = sock
        self.buffer = buffer

    def send(self, packet: Packet):
        addr = packet.addr

        if isinstance(addr, str):
            pr: ParseResult = urlparse(addr)
            addr = (pr.hostname, int(pr.port))
            assert pr.scheme == 'udp', pr

        logging.getLogger('net.trace.raw.o').debug('[%d] %s %s', len(packet.data), addr, packet.data)

        try:
            return self.fd.sendto(packet.pack(), addr)
        except socket.gaierror as e:
            if str(e).endswith('Name or service not known'):
                logging.getLogger('net.trace.raw.e').debug('%s %s', addr, e)
                return None
            else:
                raise

    def read(self) -> Iterable[Packet]:
        for x in recvfrom_helper(self.fd, self.buffer):
            logging.getLogger('net.trace.raw.i').debug('[%d] %s %s', len(x.data), x.addr,
                                                       x.data)

            addr, port = x.addr

            yield Packet(f'udp://{addr}:{port}', x.data)

    def close(self):
        self.fd.close()


class UnixTransport(UDPTransport):
    def __init__(self, url):
        Transport.__init__(self, url)

        parsed: ParseResult = urlparse(url)

        should_bind = False

        if parsed.query:
            q = parse_qs(parsed.query, keep_blank_values=True)

            if 'bind' in q:
                should_bind = True

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        if should_bind:
            sock.bind(parsed.path)
            sock.listen(1)
            sock.settimeout(0)
        else:
            sock.connect(parsed.path)

        self.fd: socket = sock


TRANSPORT_MAP: Dict[str, Type[Transport]] = {
    'udp': UDPTransport,
    'unix': UnixTransport,
}


class RPCPacketRaw(NamedTuple):
    addr: Origin
    packet: RPCPacket


RPCTransportStackMatcher = Callable[[RPCPacketRaw], bool]
RPCTransportStackProcessor = Callable[[RPCPacketRaw], None]


class RPCTransportStackEntry(NamedTuple):
    matcher: RPCTransportStackMatcher
    processor: RPCTransportStackProcessor


class RPCTransportStack:
    def __init__(self, transports: List[Transport]):
        self.transports = transports
        self.stack: List[RPCTransportStackEntry] = []

    def wait(self):
        return [x.fd for x in self.transports]

    def recv(self, polled_flags: Optional[List[bool]] = None):
        if polled_flags is None:
            polled_flags = [True for _ in self.transports]

        if not polled_flags:
            return []

        for has_data, transport in zip(polled_flags, self.transports):
            if not has_data:
                continue

            for raw_packet in transport.read():
                packet = RPCPacket.unpack(raw_packet.data)

                logging.getLogger('net.trace.pkt.i').debug('%s %s', raw_packet.addr, packet)

                for idx in range(len(self.stack) - 1, -1, -1):
                    pass

                raw_rpc_packet = RPCPacketRaw(raw_packet.addr, packet)

                for st in self.stack[::-1]:
                    if st.matcher(raw_rpc_packet):
                        st.processor(raw_rpc_packet)
                        break
                else:
                    logging.getLogger('net.trace.pkt.u').debug('%s %s', raw_packet.addr, packet)

    def send(self, packet: RPCPacketRaw):
        # todo: think about how this would work with RPC groups assigned to different ports

        logging.getLogger('net.trace.pkt.o').debug('%s %s', packet.addr, packet.packet)

        raw_packet = Packet(packet.addr, packet.packet.pack())

        self.transports[0].send(raw_packet)

    def push(self, fn: RPCTransportStackMatcher, processor: RPCTransportStackProcessor):
        self.stack.append(RPCTransportStackEntry(fn, processor))

    def pop(self):
        x = self.stack.pop()


def select_helper(fds, max_wait: float) -> Optional[List[bool]]:
    rd, _, er = select.select(fds, [], fds, max_wait)

    r = set(rd + er)

    return [fd in r for fd in fds]



class SerdeException(Exception):
    def __init__(self, val=None, code=None, par=None, **kwargs):
        self.val = val
        self.code = code
        self.parent = par
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.kwargs = kwargs
        super().__init__(val, code, kwargs, par)

    def __str__(self) -> str:
        return f'{self.__class__.__name__}(Val={self.val} Code={self.code} KwArgs={self.kwargs} Par={self.parent})'

    def resolve(self):
        r = self

        while r.parent is not None and isinstance(r.parent, SerdeException):
            r = r.parent

        return r



import base64
import datetime
import inspect
import logging
import sys
import uuid
from enum import Enum
from inspect import FullArgSpec

from xrpc.generic import build_generic_context as _build_generic_context
from xrpc.serde.error import SerdeException


def build_generic_context(*args):
    _, r = _build_generic_context(*args)
    return r


if sys.version_info >= (3, 7):
    from typing import Any, ForwardRef, Optional, Union, List, Dict, Callable, \
        NamedTuple, Tuple, TypeVar, Iterable
else:
    from typing import Any, _ForwardRef, Optional, Union, List, Dict, Callable, \
        NamedTuple, Tuple, TypeVar, Iterable
from dataclasses import is_dataclass, fields, dataclass

from xrpc.util import time_parse
from xrpc.serde.abstract import SerdeType, SerdeInst, SerdeNode, DESER, SER, SerdeTypeDeserializer, SerdeTypeSerializer, \
    SerdeStepContext

if sys.version_info >= (3, 7):
    FR = ForwardRef
else:
    FR = _ForwardRef


class ForwardRefSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        return isinstance(t, FR)

    def norm(self, i: SerdeInst, t: FR, ctx: SerdeStepContext):
        if sys.version_info >= (3, 7):
            t = t._evaluate(ctx.mod.__dict__, ctx.mod.__dict__)
        else:
            t = t._eval_type(ctx.mod.__dict__, ctx.mod.__dict__)
        t = i.norm(t, ctx)
        return t

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return i.step(i.norm(t, ctx), ctx)


class UnionNext(Exception):
    pass


class UnionDeserializer(SerdeTypeDeserializer):
    def __init__(self, parent: 'UnionSerde', t: Any, deps: List[DESER]):
        super().__init__(parent, t, deps)

        assert hasattr(t, '__args__')

        self.ordered = {k: i for i, (k, v) in enumerate(parent.extract_ordered(t))}

    def __call__(self, val: Union[list, dict]) -> Any:
        idx, val = val

        dep_idx = self.ordered.get(idx)

        if dep_idx is None:
            raise SerdeException(val, 'u_idx', idx)

        dep = self.deps[dep_idx]

        try:
            return dep(val)
        except Exception as e:
            raise SerdeException(val, 'u_deps', idx=dep_idx, par=e)


class UnionSerializer(UnionDeserializer):
    def __call__(self, val):
        desc = None

        desc = self.parent.descriptor_fn(True, val)
        nv = desc, val
        try:
            return desc, super().__call__(nv)
        except Exception as e:
            raise SerdeException(val, 'union_ser', desc=desc, par=e)


def descriptor_classname(is_inst, t):
    if not is_inst:
        return t.__name__
    else:
        return t.__class__.__name__


def is_union(t):
    if sys.version_info >= (3, 7):
        return getattr(t, '__origin__', None) is Union
    else:
        return t.__class__.__name__ == '_Union'


class UnionSerde(SerdeType):
    cls_deserializer = UnionDeserializer
    cls_serializer = UnionSerializer

    def __init__(self, descriptor_fn=None):
        if descriptor_fn is None:
            descriptor_fn = descriptor_classname
        self.descriptor_fn = descriptor_fn
        super().__init__()

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        r = is_union(t)

        if r:
            sub_args = t.__args__

            if sub_args[-1] == type(None):
                return False
            elif len(sub_args) == 1:
                return False
            else:
                return True
        else:
            return False

    def norm(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> Any:
        st = tuple(i.norm(x, ctx) for x in t.__args__)
        return Union[st]

    def extract_ordered(self, t: Any):
        vt = t.__args__

        ordered = [(self.descriptor_fn(False, x), i) for i, x in enumerate(vt)]
        ordered = sorted(ordered)
        seen = set()

        for n, idx in ordered:
            if n in seen:
                assert False, (n, 'Could not unionize types with the same name')
            seen.add(n)

        return tuple((n, vt[idx]) for n, idx in ordered)

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        ordered = self.extract_ordered(Union[tuple(i.norm(x, ctx) for x in t.__args__)])

        return SerdeNode(Union[tuple(x for _, x in ordered)], [t for _, t in ordered], ctx)


class OptionalDeserializer(SerdeTypeDeserializer):
    def __call__(self, val: Union[list, dict]) -> Any:
        fn = self.deps[0]
        try:
            if val is None:
                return None
            else:
                return fn(val)
        except Exception as e:
            raise SerdeException(val, 'opt', fn=fn, par=e)


class OptionalSerde(SerdeType):
    cls_deserializer = OptionalDeserializer
    cls_serializer = OptionalDeserializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        r = is_union(t)

        if r:
            sub_args = t.__args__

            if sub_args[-1] == type(None) and len(sub_args) == 2:
                return True
            elif len(sub_args) == 1:
                return False
            else:
                return sub_args[-1] == type(None)
        else:
            return False

    def norm(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> Any:
        *args, _ = t.__args__
        args = tuple(i.norm(x, ctx) for x in args)

        if len(args) > 1:
            return Optional[Union[args]]
        else:
            return Optional[args[0]]

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        *args, _ = t.__args__

        args = tuple(i.norm(x, ctx) for x in args)

        deps = []

        if len(args) > 1:
            deps = [Union[args]]
        else:
            deps = [args[0]]

        return SerdeNode(self.norm(i, t, ctx), deps, ctx)


class AtomDeserializer(SerdeTypeSerializer):
    def __init__(self, parent: 'SerdeType', t: Any, deps: List[SER]):
        super().__init__(parent, t, deps)
        self.mapper, *_ = [x for x in self.parent.ATOMS if t == x]

    def __call__(self, val: Any) -> Union[list, dict]:
        try:
            return self.mapper(val)
        except Exception as e:
            raise SerdeException(val, 'a_dv', t=self.mapper.__name__, par=e)


class AtomSerializer(SerdeTypeSerializer):
    def __init__(self, parent: 'SerdeType', t: Any, deps: List[SER]):
        super().__init__(parent, t, deps)
        self.mapper, *_ = [x for x in self.parent.ATOMS if t == x]

    def __call__(self, val: Any) -> Union[list, dict]:
        if not isinstance(val, self.mapper):
            raise SerdeException(val, 'a_sv', t=self.mapper.__name__)

        return val


class AtomSerde(SerdeType):
    ATOMS = (int, str, bool, float)

    cls_deserializer = AtomDeserializer
    cls_serializer = AtomSerializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        return t in self.ATOMS

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return SerdeNode(t, [])


class BytesDeserializer(SerdeTypeDeserializer):
    def __call__(self, val: Union[list, dict, str]) -> Any:
        return base64.b64decode(val.encode())


class BytesSerializer(SerdeTypeSerializer):
    def __call__(self, val: Any) -> Union[list, dict]:
        return base64.b64encode(val).decode()


class BytesSerde(SerdeType):
    cls_deserializer = BytesDeserializer
    cls_serializer = BytesSerializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if inspect.isclass(t):
            return issubclass(t, bytes)
        else:
            return False

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return SerdeNode(t, [])


class TypeVarSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        return isinstance(t, TypeVar)

    def norm(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> Any:
        if t in ctx.generic_vals:
            return ctx.generic_vals[t]
        else:
            raise ValueError(f'Only instantated generics are allowed for serialization {t} {ctx}')

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return i.step(self.norm(i, t, ctx), ctx)


class NoneSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if t == type(None):
            return True
        if t is None:
            return True
        return issubclass(type(t), type(None))

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return SerdeNode(t, [])

    def deserializer(self, t: Any, deps: List[DESER]) -> DESER:
        def deser(val):
            if val is not None:
                raise ValueError(f'Must be None: {val}')
            return None

        return deser

    def serializer(self, t: Any, deps: List[DESER]) -> DESER:
        return lambda val: None


class UUIDSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if inspect.isclass(t):
            return issubclass(t, uuid.UUID)
        else:
            return False

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return SerdeNode(t, [])

    def deserializer(self, t: Any, deps: List[DESER]) -> DESER:
        return lambda val: uuid.UUID(hex=val)


ISO8601 = '%Y-%m-%dT%H:%M:%S.%f'


class DateTimeSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if inspect.isclass(t):
            return issubclass(t, datetime.datetime)
        else:
            return False

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return SerdeNode(t, [])

    def deserializer(self, t: Any, deps: List[DESER]) -> DESER:
        return lambda val: time_parse(val, ISO8601)

    def serializer(self, t: Any, deps: List[DESER]) -> DESER:
        return lambda val: format(val, ISO8601)


class ListDeserializer(SerdeTypeDeserializer):
    cls_coll = list

    def __call__(self, val: Union[list, dict]) -> Any:
        self.d, *_ = self.deps
        return self.cls_coll(self.d(v) for v in val)


class ListSerializer(ListDeserializer, SerdeTypeSerializer):
    pass


class ListSerde(SerdeType):
    cls_deserializer = ListDeserializer
    cls_serializer = ListSerializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        t, _ = _build_generic_context(t, ctx)

        if sys.version_info >= (3, 7):
            if hasattr(t, '__origin__'):
                return t.__origin__ is list

        if inspect.isclass(t):
            return issubclass(t, List)
        else:
            return False

    def norm(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> Any:
        _, ctx = _build_generic_context(t, ctx)
        st, = t.__args__
        st = i.norm(st, ctx)

        return List[st]

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        t = i.norm(t, ctx)
        return SerdeNode(t, [t.__args__[0]], ctx=ctx)


class TupleDeserializer(ListDeserializer):
    cls_coll = tuple


class TupleSerializer(TupleDeserializer, SerdeTypeSerializer):

    def __call__(self, val: Union[list, dict]) -> Any:
        return list(super().__call__(val))


class TupleSerde(SerdeType):
    cls_deserializer = TupleDeserializer
    cls_serializer = TupleSerializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        t, _ = _build_generic_context(t, ctx)

        if sys.version_info >= (3, 7):
            if hasattr(t, '__origin__'):
                return t.__origin__ is tuple

        if inspect.isclass(t):
            return issubclass(t, Tuple)
        else:
            return False

    def norm(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> Any:
        _, ctx = _build_generic_context(t, ctx)
        args = tuple(i.norm(st, ctx) for st in t.__args__)

        return Tuple[args]

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        t = i.norm(t, ctx)
        return SerdeNode(t, [x for x in t.__args__], ctx=ctx)


class DictSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if sys.version_info >= (3, 7):
            is_gen = hasattr(t, '__origin__')

            if is_gen:
                t = t.__origin__

        if inspect.isclass(t):
            return issubclass(t, Dict)
        else:
            return False

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        kt, vt = t.__args__
        kt, vt = i.norm(kt, ctx), i.norm(vt, ctx)
        return SerdeNode(Dict[kt, vt], [kt, vt])

    def deserializer(self, t: Any, deps: List[DESER]) -> DESER:
        def deser_dict(val):
            r = {}
            try:
                for k, v in val.items():
                    r[deps[0](k)] = deps[1](v)
            except BaseException as e:
                raise NotImplementedError(f'{t} {k} {v} {e}')
            return r

        return deser_dict

    def serializer(self, t: Any, deps: List[SER]) -> SER:
        def ser_dict(val):
            ks, vs = deps
            return {ks(k): vs(v) for k, v in val.items()}

        return ser_dict


class EnumSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if inspect.isclass(t):
            return issubclass(t, Enum)
        else:
            return False

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return SerdeNode(t, [])

    def deserializer(self, t: Any, deps: List[DESER]) -> DESER:
        assert False, ''


def build_obj_module(obj):
    m = obj.__module__

    return sys.modules[m]


class NamedTupleDeserializer(SerdeTypeDeserializer):
    def __init__(self, parent: 'NamedTupleSerde', t: Any, deps: List[DESER]):
        super().__init__(parent, t, deps)
        self.fields = [f for f, _ in t._field_types.items()]

    def __call__(self, val: Union[list, dict]) -> Any:
        r = {}

        i, f = None, None

        try:

            for i, f in enumerate(self.fields):
                # either way (even if a field is non-optional), that field must not accept None as it's argument
                r[f] = self.deps[i](val.get(f))
                # r[f] = self.deps[i](val[f])
        except Exception as e:
            raise SerdeException(val, 'u_sf', idx=i, field=f, par=e)

        try:

            return self.t(**r)
        except Exception as e:
            raise SerdeException(val, 'u_inst', fields=r, par=e)


class NamedTupleSerializer(SerdeTypeSerializer):
    def __init__(self, parent: 'SerdeType', t: Any, deps: List[SER]):
        super().__init__(parent, t, deps)
        self.fields = [f for f, _ in t._field_types.items()]

    def __call__(self, val: Any) -> Union[list, dict]:
        return {v: self.deps[k](getattr(val, v)) for k, v in enumerate(self.fields)}


class NamedTupleSerde(SerdeType):
    cls_deserializer = NamedTupleDeserializer
    cls_serializer = NamedTupleSerializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if sys.version_info >= (3, 7):
            is_gen = hasattr(t, '__origin__')

            if is_gen:
                t = t.__origin__

        return hasattr(t, '_field_types')

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        xt = t

        if sys.version_info >= (3, 7):
            is_gen = hasattr(t, '__origin__')

            if is_gen:
                xt = t.__origin__

        ctx = SerdeStepContext(mod=build_obj_module(t))
        return SerdeNode(t, [i.norm(st, ctx) for f, st in xt._field_types.items()], ctx=ctx)


class DataclassDeserializer(SerdeTypeDeserializer):
    def __init__(self, parent: 'SerdeType', t: Any, deps: List[DESER]):
        super().__init__(parent, t, deps)

        if sys.version_info >= (3, 7):
            is_gen = hasattr(t, '__origin__')

            if is_gen:
                t = t.__origin__

        self.fields = [f.name for f in sorted(fields(t), key=lambda x: x.name)]

    def __call__(self, val: Union[list, dict]) -> Any:
        return NamedTupleDeserializer.__call__(self, val)


class DataclassSerializer(SerdeTypeDeserializer):
    def __init__(self, parent: 'SerdeType', t: Any, deps: List[DESER]):
        super().__init__(parent, t, deps)

        if sys.version_info >= (3, 7):
            is_gen = hasattr(t, '__origin__')

            if is_gen:
                t = t.__origin__

        self.fields = [f.name for f in sorted(fields(t), key=lambda x: x.name)]

    def __call__(self, val: Union[list, dict]) -> Any:
        return NamedTupleSerializer.__call__(self, val)


class DataclassSerde(SerdeType):
    cls_deserializer = DataclassDeserializer
    cls_serializer = DataclassSerializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        if sys.version_info >= (3, 7):
            is_gen = hasattr(t, '__origin__')

            if is_gen:
                t = t.__origin__
        return is_dataclass(t)

    def norm(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> Any:
        t, ctx = _build_generic_context(t, ctx)
        return t

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        # todo: here, we may need to instantiate the generic items
        t, ctx = _build_generic_context(t, ctx)

        xt = t

        if sys.version_info >= (3, 7):
            is_gen = hasattr(t, '__origin__')

            if is_gen:
                xt = t.__origin__

        logging.getLogger(__name__ + '.DataclassSerde.step').debug('%s %s', t, ctx)

        return SerdeNode(i.norm(t, ctx), [i.norm(f.type, ctx) for f in sorted(fields(xt), key=lambda x: x.name)], ctx)


class CallableArgsWrapper(NamedTuple):
    method: bool
    name: str
    spec: FullArgSpec
    cls: Optional[Any] = None

    @classmethod
    def from_func(self, fn: Callable):
        return CallableArgsWrapper(inspect.ismethod(fn), f'{fn.__module__}.{fn.__name__}', inspect.getfullargspec(fn))

    @classmethod
    def from_func_cls(self, cls: Any, fn: Callable):
        return CallableArgsWrapper(True, f'{fn.__module__}.{fn.__name__}', inspect.getfullargspec(fn), cls)

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            return super().__eq__(other)
        else:
            return False

    def __hash__(self) -> int:
        r = (
            'CallableArgsWrapper',
            self.name,
            self.method,
            tuple(self.spec.args),
            self.spec.varargs,
            self.spec.varkw,
            self.spec.defaults,
            tuple(self.spec.kwonlyargs),
            None if self.spec.kwonlydefaults is None else tuple(sorted(self.spec.kwonlydefaults.items())),
            None if self.spec.annotations is None else tuple(sorted(self.spec.annotations.items())),
            self.cls
        )
        return hash(r)


class CallableRetWrapper(NamedTuple):
    method: bool
    name: str
    spec: FullArgSpec
    cls: Optional[Any] = None

    @classmethod
    def from_func(self, fn: Callable):
        return CallableRetWrapper(inspect.ismethod(fn), f'{fn.__module__}.{fn.__name__}', inspect.getfullargspec(fn))

    @classmethod
    def from_func_cls(self, cls: Any, fn: Callable):
        return CallableRetWrapper(True, f'{fn.__module__}.{fn.__name__}', inspect.getfullargspec(fn), cls)

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            return super().__eq__(other)
        else:
            return False

    def __hash__(self) -> int:
        r = (
            'CallableRetWrapper',
            self.name,
            self.method,
            tuple(self.spec.args),
            self.spec.varargs,
            self.spec.varkw,
            self.spec.defaults,
            tuple(self.spec.kwonlyargs),
            None if self.spec.kwonlydefaults is None else tuple(sorted(self.spec.kwonlydefaults.items())),
            None if self.spec.annotations is None else tuple(sorted(self.spec.annotations.items())),
            self.cls
        )
        return hash(r)


def build_types(spec: FullArgSpec, is_method=False, allow_missing=False):
    missing_args = []
    map = {}

    args = spec.args

    def get_annot(name):
        if name not in spec.annotations:
            missing_args.append(name)
            return None
        return spec.annotations[name]

    if is_method:
        args = args[1:]

    for arg in args:
        map[arg] = get_annot(arg)

    if spec.varargs:
        map[ARGS_VAR] = get_annot(spec.varargs)

    if spec.varkw:
        map[ARGS_KW] = get_annot(spec.varkw)

    if spec.kwonlyargs:
        for arg in spec.kwonlyargs:
            map[arg] = get_annot(arg)

    if spec.annotations and 'return' in spec.annotations:
        map[ARGS_RET] = spec.annotations['return']
    else:
        map[ARGS_RET] = None

    if not allow_missing and len(missing_args):
        raise NotImplementedError(
            f'Can not find annotations for arguments named: `{missing_args}`')

    return map


EMPTY_DEFAULT = object()


class ArgumentsException(Exception):
    def __init__(
            self,
            is_pos=False,
            is_pos_many=False,
            name_missing: Optional[List[str]] = None,
            argument_required: Optional[List[str]] = None,
    ):
        self.is_pos = is_pos
        self.is_pos_many = is_pos_many
        self.name_missing = name_missing
        self.argument_required = argument_required
        super().__init__(is_pos, name_missing, argument_required)

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.is_pos}, {self.is_pos_many}, {self.name_missing}, {self.argument_required})'


@dataclass
class PairSpecValue:
    name: str
    name_idx: Optional[Union[str, int]]
    arg: Any
    """Given argument to the function"""
    default_arg: Any


@dataclass
class PairSpec:
    spec: FullArgSpec
    is_method: bool

    def __call__(self, *args, **kwargs) -> Iterable[PairSpecValue]:
        spec = self.spec
        is_method = self.is_method
        # pair function arguments to their names

        # we need to "eat" the arguments correctly (and then map them to something else).
        map_args = list(spec.args)
        map_args_defaults = list(spec.defaults or [])
        map_args_defaults = [EMPTY_DEFAULT] * (len(map_args) - len(map_args_defaults)) + map_args_defaults

        kwonlyargs = list(spec.kwonlyargs) if spec.kwonlyargs else []
        kwonlydefaults = spec.kwonlydefaults or {}

        mapped_args = list()

        if is_method:
            assert map_args[0] == 'self'
            map_args = map_args[1:]

        vararg_ctr = 0

        names_missing = []

        for arg in args:
            if len(map_args):
                curr_arg, map_args = map_args[0], map_args[1:]
                curr_default_arg, map_args_defaults = map_args_defaults[0], map_args_defaults[1:]

                yield PairSpecValue(curr_arg, None, arg, curr_default_arg)

                mapped_args.append(curr_arg)
            elif spec.varargs:
                yield PairSpecValue(ARGS_VAR, vararg_ctr, arg, None)
                vararg_ctr += 1

            else:
                raise ArgumentsException(is_pos=True, is_pos_many=True, name_missing=[arg])

        matched_kwds = []

        for kwarg_name, kwarg_val in kwargs.items():
            has_matched = False
            default_value = EMPTY_DEFAULT

            if kwarg_name in map_args:
                map_args_idx = map_args.index(kwarg_name)

                has_matched = True
                default_value = map_args_defaults[map_args_idx]

                map_args = map_args[:map_args_idx] + map_args[map_args_idx + 1:]
                map_args_defaults = map_args_defaults[:map_args_idx] + map_args_defaults[map_args_idx + 1:]

            elif kwarg_name in kwonlyargs:
                kwonlyargs.remove(kwarg_name)
                has_matched = True
                default_value = kwonlydefaults.get(kwarg_name, EMPTY_DEFAULT)

            if not has_matched:
                if spec.varkw:
                    assert kwarg_name not in mapped_args, kwarg_name
                    yield PairSpecValue(ARGS_KW, kwarg_name, kwarg_val, None)
                else:
                    names_missing.append(kwarg_name)
                    continue
            else:
                yield PairSpecValue(kwarg_name, None, kwarg_val, default_value)

            matched_kwds.append(kwarg_name)

        reqd = []
        is_pos = True

        for v in kwonlyargs:
            if v in kwonlydefaults:
                is_pos = False
                yield PairSpecValue(ARGS_KW, v, EMPTY_DEFAULT, kwonlydefaults[v])
            else:
                reqd.append(v)

        for x, y in zip(map_args, map_args_defaults):
            if y is EMPTY_DEFAULT:
                reqd.append(y)
            else:
                yield ARGS_KW, x, EMPTY_DEFAULT, y

        if len(reqd):
            raise ArgumentsException(is_pos=is_pos, argument_required=reqd)


def pair_spec(spec: FullArgSpec, is_method, *args, **kwargs):
    return PairSpec(spec, is_method)(*args, **kwargs)


ARGS_VAR = '$var'
ARGS_KW = '$kw'
ARGS_RET = '$ret'


class CallableArgsSerde(SerdeType):
    # fullargspect -> Type (or a function)
    # args_kwargs -> INPUT

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        return isinstance(t, CallableArgsWrapper)

    def _build_args(self, t: CallableArgsWrapper):
        at = t.spec

        missing_args = []
        map = {}

        args = at.args

        def get_annot(name):
            if name not in at.annotations:
                missing_args.append(name)
                return None
            return at.annotations[name]

        if t.method:
            args = args[1:]

        for arg in args:
            map[arg] = get_annot(arg)

        if at.varargs:
            map[ARGS_VAR] = get_annot(at.varargs)

        if at.varkw:
            map[ARGS_KW] = get_annot(at.varkw)

        if at.kwonlyargs:
            for arg in at.kwonlyargs:
                map[arg] = get_annot(arg)

        if len(missing_args):
            raise NotImplementedError(
                f'Function `{t.method}` `{t.name}` not find annotations for arguments named: `{missing_args}`')

        return map

    def step(self, i: SerdeInst, t: CallableArgsWrapper, ctx: SerdeStepContext) -> SerdeNode:
        if t.cls:
            ctx = build_generic_context(t.cls, ctx)
            ctx = ctx.merge(SerdeStepContext(mod=build_obj_module(t.cls)))

        types = sorted(self._build_args(t).items(), key=lambda x: x[0])

        types = [i.norm(x, ctx) for _, x in types]

        return SerdeNode(t, types, ctx)

    def deserializer(self, t: CallableArgsWrapper, deps: List[DESER]) -> DESER:
        at = t.spec
        map = self._build_args(t)
        map = sorted(map.items(), key=lambda x: x[0])
        map = zip(map, enumerate(deps))
        map = {k: i for (k, _), (i, _) in map}

        def get_map(name):
            return deps[map[name]]

        def callable_deserializer(val: list) -> list:
            # given a val, pair the values with the types

            val: Tuple[List[Any], Dict[str, Any]]
            args, kwargs = val

            # we need to "eat" the arguments correctly (and then map them to something else).

            map_args = t.spec.args

            if t.method:
                map_args = map_args[1:]

            r_args, r_kwargs = tuple(), {}

            for arg in args:
                if len(map_args):
                    curr_arg, map_args = map_args[0], map_args[1:]

                    r_args = r_args + (get_map(curr_arg)(arg),)
                elif t.spec.varargs:
                    r_args = r_args + (get_map(ARGS_VAR)(arg),)
                else:
                    raise ValueError(f'Could not find mapping for argument `{arg}`')

            for kwarg_name, kwarg_val in kwargs.items():
                if kwarg_name not in map:
                    if at.varkw:
                        r_kwargs[kwarg_name] = get_map(ARGS_KW)(kwarg_val)
                    else:
                        raise ValueError(f'Function does not accept `{kwarg_name}`')
                else:
                    r_kwargs[kwarg_name] = get_map(kwarg_name)(kwarg_val)

            return [r_args, r_kwargs]

        return callable_deserializer

    def serializer(self, t: Any, deps: List[SER]) -> SER:
        at = t.spec
        map = self._build_args(t)
        map = sorted(map.items(), key=lambda x: x[0])
        map = zip(map, enumerate(deps))
        map = {k: i for (k, _), (i, _) in map}

        def get_map(name):
            return deps[map[name]]

        def callable_serializer(val: SER) -> list:
            val: Tuple[List[Any], Dict[str, Any]]
            args, kwargs = val

            map_args = t.spec.args

            if t.method:
                map_args = map_args[1:]

            r_args, r_kwargs = tuple(), {}

            for arg in args:
                if len(map_args):
                    curr_arg, map_args = map_args[0], map_args[1:]

                    r_args = r_args + (get_map(curr_arg)(arg),)
                elif t.spec.varargs:
                    r_args = r_args + (get_map(ARGS_VAR)(arg),)
                else:
                    raise ValueError(f'Could not find mapping for argument `{arg}`')

            for kwarg_name, kwarg_val in kwargs.items():
                if kwarg_name not in map:
                    if at.varkw:
                        r_kwargs[kwarg_name] = get_map(ARGS_KW)(kwarg_val)
                    else:
                        raise ValueError(f'Function does not accept `{kwarg_name}`')
                else:
                    r_kwargs[kwarg_name] = get_map(kwarg_name)(kwarg_val)

            return [r_args, r_kwargs]

        return callable_serializer


class CallableRetSerde(SerdeType):
    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        return isinstance(t, CallableRetWrapper)

    def step(self, i: SerdeInst, t: CallableRetWrapper, ctx: SerdeStepContext) -> SerdeNode:
        if t.cls:
            ctx = build_generic_context(t.cls, ctx)
            ctx = ctx.merge(SerdeStepContext(mod=build_obj_module(t.cls)))

        RET = 'return'

        dt = None

        if RET in t.spec.annotations:
            dt = i.norm(t.spec.annotations[RET], ctx)
        return SerdeNode(t, [dt], ctx)

    def deserializer(self, t: CallableArgsWrapper, deps: List[DESER]) -> DESER:
        def callable_ret_deser(val):
            dep = deps[0]

            try:
                return dep(val)
            except Exception as e:
                raise SerdeException(val=val, code='ex', par=e)

        return callable_ret_deser

    def serializer(self, t: CallableArgsWrapper, deps: List[SER]) -> SER:
        def callable_ret_ser(val):
            dep = deps[0]

            try:
                return dep(val)
            except Exception as e:
                raise SerdeException(val=val, code='ex', par=e)

        return callable_ret_ser



import heapq
from collections import deque
from datetime import datetime
from typing import TypeVar, Generic, List, Optional, Deque

from xrpc.util import time_now


class MutableInt:
    def __init__(self, state: int = 0):
        self.state = int(state)

    def __set__(self, instance, value):
        pass

    def __iadd__(self, other):
        self.state += other

    def __isub__(self, other):
        self.state -= other

    def __le__(self, other):
        return self.state <= other

    def set(self, state: int):
        self.state = int(state)

    def reduce(self, x=1):
        self.state -= x

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.state})'


class MutableDateTime:
    def __init__(self, t: datetime):
        self.t = t

    @classmethod
    def now(cls):
        return MutableDateTime(time_now())

    def get(self) -> datetime:
        return self.t

    def set(self, t: datetime) -> datetime:
        r = self.t
        self.t = t
        return r


T = TypeVar('T')
H = TypeVar('H')


class Queue(Generic[T, H]):
    def __init__(self, initial: Optional[List[T]] = None):
        self.h: H = self._init(initial)

    def _init(self, initial: List[T]) -> H:
        raise NotImplementedError()

    def push(self, val: T):
        raise NotImplementedError()

    def pop(self) -> T:
        raise NotImplementedError()

    def peek(self) -> Optional[T]:
        raise NotImplementedError()


class BinaryQueue(Queue[T, Deque[T]]):
    def _init(self, initial: Optional[List[T]]) -> Deque[T]:
        if initial:
            return deque(sorted(initial))
        else:
            return deque()

    def push(self, val: T):
        if len(self.h) == 0:
            self.h.append(val)
            return

        min_idx = 0
        max_idx = len(self.h) - 1

        while min_idx != max_idx and min_idx < max_idx - 1:
            mid_idx = (min_idx + max_idx) // 2

            mid_val = self.h[mid_idx]

            prev_range = (min_idx, max_idx)

            if mid_val == val:
                break
            elif mid_val > val:
                max_idx = mid_idx - 1
            else:
                min_idx = mid_idx + 1

            assert prev_range != (min_idx, max_idx), f'Ranges must always differ, {prev_range}, {(min_idx, max_idx)}'

        if min_idx == max_idx:
            if self.h[min_idx] < val:
                self.h.insert(min_idx + 1, val)
            else:
                self.h.insert(min_idx, val)
            return
        else:
            if self.h[min_idx] > val:
                self.h.insert(min_idx, val)
            elif self.h[max_idx] < val:
                self.h.insert(max_idx + 1, val)
            else:
                self.h.insert(min_idx + 1, val)
            return

    def pop(self) -> T:
        r = self.h.popleft()
        return r

    def peek(self) -> Optional[T]:
        if len(self.h):
            return self.h[0]
        else:
            return None


class HeapQueue(Queue[T, List[T]]):
    def _init(self, initial: Optional[List[T]]) -> List[T]:
        if initial:
            initial = list(initial)
            heapq.heapify(initial)
            return initial
        else:
            return []

    def push(self, val: T):
        heapq.heappush(self.h, val)

    def pop(self) -> T:
        return heapq.heappop(self.h)

    def peek(self) -> Optional[T]:
        if len(self.h):
            return self.h[0]
        else:
            return None

    def pushpop(self, val: T):
        return heapq.heappushpop(self.h, val)

    def replace(self, val: T):
        return heapq.heapreplace(self.h, val)



import json
import logging
import os
import tempfile
from collections import deque
from inspect import getfullargspec, ismethod

import shutil
import socket
import types
from dataclasses import dataclass
from datetime import datetime
from itertools import count
from signal import SIGTERM
from subprocess import TimeoutExpired
from time import sleep
from typing import NamedTuple, Callable, Optional, Dict, Deque, TypeVar, Generic, Type, Tuple, Union

from xrpc.abstract import MutableInt
from xrpc.cli import Parsable
from xrpc.client import ClientConfig
from xrpc.const import SERVER_SERDE_INST
from xrpc.dsl import rpc, RPCType, regular, socketio, signal
from xrpc.error import HorizonPassedError, TimeoutError, TerminationException
from xrpc.logging import logging_config, LoggerSetup, logging_setup, circuitbreaker
from xrpc.popen import popen
from xrpc.runtime import service, sender
from xrpc.serde.abstract import SerdeSet, SerdeStruct
from xrpc.serde.types import build_types, ARGS_RET, PairSpec
from xrpc.transport import recvfrom_helper, Packet, Origin
from xrpc.util import time_now, signal_context


@dataclass
class BrokerConf(Parsable):
    heartbeat: float = 5.
    max_pings: int = 5
    metrics: float = 10.


@dataclass
class WorkerMetric:
    running_since: Optional[datetime]


@dataclass
class BrokerMetric:
    workers: int
    jobs_pending: int
    jobs: int
    assigned: int


NodeMetric = Union[WorkerMetric, BrokerMetric]

RequestType = TypeVar('RequestType')
ResponseType = TypeVar('ResponseType')

WorkerCallable = Callable[[RequestType], ResponseType]


class MetricCollector:
    @rpc(RPCType.Signalling)
    def metrics(self, metric: NodeMetric):
        pass


def get_func_types(fn: WorkerCallable) -> Tuple[Type[RequestType], Type[ResponseType]]:
    if not isinstance(fn, types.FunctionType):
        fn = fn.__call__

    spec = getfullargspec(fn)
    is_method = ismethod(fn)
    annot = build_types(spec, is_method, allow_missing=True)
    arg = next(PairSpec(spec, is_method)(None))

    return annot[arg.name], annot[ARGS_RET]


def worker_inst(logger_config: LoggerSetup, fn: WorkerCallable, path: str):
    def sig_handler(code, frame):
        logging.getLogger(__name__).error(f'Received {code}')
        raise KeyboardInterrupt('')

    with logging_setup(logger_config), circuitbreaker(main_logger='broker'), signal_context(handler=sig_handler):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)  # UDP

        logging.getLogger('worker_inst').debug('Binding to %s', path)
        sock.bind(path)
        sock.listen(1)

        connection, origin = sock.accept()
        logging.getLogger('worker_inst.accept').debug('%s', origin)

        # use the callable's type hints in order to serialize and deserialize parameters

        cls_req, cls_res = get_func_types(fn)

        serde = build_serde(cls_req, cls_res)

        try:
            for x in recvfrom_helper(connection, logger_name='worker_inst.net.trace.raw'):
                logging.getLogger('worker_inst.net.trace.raw.i').debug('[%d] %s %s', len(x.data), x.addr,
                                                                       x.data)
                jp: cls_req = serde.deserialize(cls_req, json.loads(x.data))

                ret = fn(jp)

                op = Packet(None, json.dumps(serde.serialize(cls_res, ret)).encode())

                logging.getLogger('worker_inst.net.trace.raw.o').debug('[%d] %s %s', len(op.data), op.addr,
                                                                       op.data)

                connection.send(op.pack())
        except KeyboardInterrupt:
            logging.getLogger('worker_inst').debug('Mildly inconvenient exit')
        finally:
            sock.close()


def build_serde(req: Type[ResponseType], res: Type[ResponseType]) -> SerdeStruct:
    a = SerdeSet.walk(SERVER_SERDE_INST, req)
    b = SerdeSet.walk(SERVER_SERDE_INST, res)

    return a.merge(b).struct(SERVER_SERDE_INST)


class Worker(Generic[RequestType, ResponseType]):
    def __init__(
            self,
            cls_req: Type[RequestType], cls_res: Type[ResponseType],
            conf: BrokerConf,
            broker_addr: Origin,
            fn: WorkerCallable[RequestType, ResponseType],
            url_metrics: Optional[str] = None
    ):
        self.cls_req = cls_req
        self.cls_res = cls_res

        self.serde = build_serde(self.cls_req, self.cls_res)

        self.conf = conf
        self.broker_addr = broker_addr
        self.url_metrics = url_metrics
        self.assigned: Optional[RequestType] = None
        self.running_since: Optional[datetime] = None

        self.dir = None
        self.dir = tempfile.mkdtemp()

        unix_url = os.path.join(self.dir, 'unix.sock')
        self.unix_url = unix_url

        self.inst = popen(worker_inst, logging_config(), fn, unix_url)

        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.socket.settimeout(1)
        self.socket.setblocking(0)

        sleep(0.3)

        for attempt in count():
            try:
                self.socket.connect(unix_url)
                break
            except:
                if attempt >= 5:
                    raise ValueError('Could not instantiate a worker')
                logging.exception('At %d', attempt)
                sleep(1)

    @rpc()
    def get_assigned(self) -> Optional[ResponseType]:
        return self.assigned

    @rpc(RPCType.Durable)
    def assign(self, pars: RequestType):
        if self.assigned is not None and pars != self.assigned:
            raise ValueError('Double assignment')

        op = Packet(self.unix_url, json.dumps(self.serde.serialize(self.cls_req, pars)).encode())

        logging.getLogger('net.trace.raw.o').debug('[%d] %s %s', len(op.data), op.addr, op.data)

        self.socket.send(op.pack())
        self.assigned = pars

        self.running_since = time_now()

    @rpc(RPCType.Repliable)
    def pid(self) -> int:
        return int(self.inst.pid)

    def is_killed(self) -> bool:
        try:
            self.inst.wait(0)
            logging.getLogger(__name__).warning('Worker process had been killed')
            return True
        except TimeoutExpired:
            return False

    def possibly_killed(self, definitely=False):
        if definitely or self.is_killed():
            self.exit()
            raise TerminationException()

    @regular()
    def heartbeat(self) -> float:
        self.possibly_killed()

        s = service(Broker[self.cls_req, self.cls_res], self.broker_addr)
        s.remind()

        return self.conf.heartbeat

    @socketio()
    def bg(self):
        try:
            for x in recvfrom_helper(self.socket):
                logging.getLogger('net.trace.raw.i').debug('[%d] %s %s', len(x.data), x.addr, x.data)

                ret = self.serde.deserialize(self.cls_res, json.loads(x.data))

                logging.getLogger('bg').debug('Returned %s', x)

                self.assigned = None

                s = service(Broker[self.cls_req, self.cls_res], self.broker_addr)

                try:
                    s.done(ret)
                except HorizonPassedError:
                    logging.getLogger('bg').exception('Seems like the broker had been killed while I was working')

                self.running_since = None
        except ConnectionAbortedError:
            self.possibly_killed(True)

        return self.socket

    @regular()
    def metrics(self) -> float:
        if self.url_metrics:
            s = service(MetricCollector, self.url_metrics)
            s.metrics(WorkerMetric(
                self.running_since
            ))
        return self.conf.metrics

    @signal()
    def exit(self):
        try:
            s = service(Broker[self.cls_req, self.cls_res], self.broker_addr, ClientConfig(timeout_total=1.))
            s.leaving()
        except TimeoutError:
            logging.getLogger('exit').error('Could not contact broker')
        self.inst.send_signal(SIGTERM)
        try:
            self.inst.wait(1)
        except TimeoutExpired:
            logging.getLogger('exit').error('Could stop worker graciously')
            self.inst.kill()
        if self.dir:
            shutil.rmtree(self.dir)
        return True


class WorkerState(NamedTuple):
    pings_remaining: MutableInt


class JobState(NamedTuple):
    created: datetime

    @classmethod
    def new(cls):
        return JobState(created=time_now())


class BrokerResult(Generic[ResponseType]):
    @rpc(RPCType.Durable)
    def finished(self, job: ResponseType):
        logging.getLogger('finished').warning('unused %s', job)


class BrokerEntry(Generic[ResponseType]):
    @rpc(RPCType.Durable)
    def assign(self, pars: RequestType):
        pass


class Broker(Generic[RequestType, ResponseType], BrokerEntry[ResponseType]):
    def __init__(
            self,
            cls_req: Type[RequestType], cls_res: Type[ResponseType],
            conf: BrokerConf,
            url_results: Optional[str] = None,
            url_metrics: Optional[str] = None
    ):
        self.cls_req = cls_req
        self.cls_res = cls_res

        self.conf = conf
        self.url_results = url_results
        self.url_metrics = url_metrics

        self.workers: Dict[Origin, WorkerState] = {}

        self.jobs: Dict[RequestType, JobState] = {}
        self.jobs_pending: Deque[RequestType] = deque()

        self.workers_jobs: Dict[Origin, RequestType] = {}

    def job_new(self, pars: RequestType):
        logging.getLogger('job_new').debug('%s', pars)

        self.jobs[pars] = JobState.new()
        self.jobs_pending.append(pars)

        self.jobs_try_assign()

    def job_resign(self, k: Origin):
        j = self.workers_jobs[k]

        del self.workers_jobs[k]

        self.jobs_pending.appendleft(j)

    def jobs_try_assign(self):
        free_workers = list(set(self.workers.keys()) - set(self.workers_jobs.keys()))

        while len(free_workers) and len(self.jobs_pending):
            pars = self.jobs_pending.popleft()
            wrkr = free_workers.pop()

            s = service(Worker[self.cls_req, self.cls_res], wrkr, ClientConfig(timeout_total=1.))

            try:
                s.assign(pars)
            except TimeoutError:
                logging.getLogger('jobs_try_assign').error('Timeout %s', wrkr)
                self.jobs_pending.appendleft(pars)
                continue
            else:
                logging.getLogger('jobs_try_assign').debug('%s %s', wrkr, pars)
                self.workers_jobs[wrkr] = pars

    def worker_new(self, k: Origin):
        logging.getLogger('worker_new').debug('%s', k)

        self.workers[k] = WorkerState(MutableInt(self.conf.max_pings))

        self.jobs_try_assign()

    def worker_lost(self, k: Origin):
        logging.getLogger('worker_lost').debug('%s', k)

        if k in self.workers_jobs:
            self.job_resign(k)

        del self.workers[k]

        self.jobs_try_assign()

    def worker_done(self, w: Origin):
        if w not in self.workers:
            logging.getLogger('job_done').warning('Not registered %s', w)
            return

        if w not in self.workers_jobs:
            logging.getLogger('job_done').warning('Worker is not assigned any jobs %s', w)
            return

        j = self.workers_jobs[w]

        del self.jobs[j]
        del self.workers_jobs[w]

        self.jobs_try_assign()

    @rpc()
    def stats(self) -> Tuple[int]:
        return len(self.workers),

    @rpc(RPCType.Durable)
    def assign(self, pars: RequestType):
        """
        Assign a job to the broker
        :param pars:
        :return:
        """
        if pars not in self.jobs:
            self.job_new(pars)
        else:
            logging.getLogger('assign').warning('Job is still working %s', pars)

    @rpc(RPCType.Durable)
    def done(self, jr: ResponseType):
        if self.url_results:
            try:
                s = service(BrokerResult[self.cls_res], self.url_results)

                s.finished(jr)
            except HorizonPassedError:
                pass
        else:
            logging.getLogger('done').error('Return type not used %s', jr)

        # todo 1) keep a log of completed jobs
        # todo 2) reply to sender
        # todo 3) send downstream

        self.worker_done(sender())

    @rpc(RPCType.Durable)
    def leaving(self):
        s = sender()

        if s in self.workers:
            self.worker_lost(s)

    @rpc(RPCType.Signalling)
    def remind(self):
        s = sender()

        if s not in self.workers:
            self.worker_new(s)
        else:
            self.workers[s].pings_remaining.set(self.conf.max_pings)

    @regular()
    def gc(self) -> float:
        for k in list(self.workers.keys()):
            self.workers[k].pings_remaining.reduce(1)

            logging.getLogger('gc').debug('%s %s', k, self.workers[k])

            if self.workers[k].pings_remaining <= 0:
                self.worker_lost(k)

        return self.conf.heartbeat

    @regular()
    def metrics(self) -> float:
        # how to we allow for reflection in the API?

        if self.url_metrics:
            s = service(MetricCollector, self.url_metrics)
            s.metrics(BrokerMetric(
                len(self.workers),
                len(self.jobs_pending),
                len(self.jobs),
                len(self.workers_jobs)
            ))

        return self.conf.metrics

    @signal()
    def exit(self):
        return True



import logging
import threading
from typing import Optional, NamedTuple, Callable, Dict, TypeVar, Type, Any

from xrpc.client import ClientConfig, ServiceWrapper
from xrpc.net import RPCKey
from xrpc.service import ServiceDefn
from xrpc.transport import Origin, RPCTransportStack

RUNTIME_TL = threading.local()
CTX_NAME = 'rpc_config'
CACHE_NAME = 'rpc_cache'


class ExecutionContext(NamedTuple):
    transport_stack: RPCTransportStack
    origin: Optional[Origin] = None
    key: Optional[RPCKey] = None
    ret: Optional[Callable[[Any], None]] = None

    def exec(self, __origin, __fn: Callable, *args, **kwargs):
        is_ok = False
        r = None

        logging.getLogger(f'xrpc.trace.e.{__origin}').debug('Name=%s %s %s %s %s', __fn.__name__, is_ok, args, kwargs, r)
        try:
            context_set(self)

            r = __fn(*args, **kwargs)
            is_ok = True
            return r
        finally:
            logging.getLogger(f'xrpc.trace.x.{__origin}').debug('Name=%s %s %s %s %s', __fn.__name__, is_ok, args, kwargs, r)
            context_set(None)


def context_set(config: Optional[ExecutionContext]):
    if sender:
        setattr(RUNTIME_TL, CTX_NAME, config)
    elif hasattr(RUNTIME_TL, CTX_NAME):
        delattr(RUNTIME_TL, CTX_NAME)


def cache_get() -> Dict[Type, ServiceDefn]:
    if not hasattr(RUNTIME_TL, CACHE_NAME):
        nd = {}
        setattr(RUNTIME_TL, CACHE_NAME, nd)

    return getattr(RUNTIME_TL, CACHE_NAME)


def context() -> ExecutionContext:
    return getattr(RUNTIME_TL, CTX_NAME)


def sender() -> Origin:
    return context().origin


def reply(ret: Any):
    ctx = context()

    ctx.ret(ret)


T = TypeVar('T')


def service(obj: Type[T], addr: Origin, conf: ClientConfig = ClientConfig()) -> T:
    ctx = context()

    service_cache = cache_get()

    if obj in service_cache:
        defn = service_cache[obj]
    else:
        defn = ServiceDefn.from_obj(obj)
        service_cache[obj] = defn

    return ServiceWrapper(defn, conf, ctx.transport_stack, addr)



import logging
import random
from time import sleep
from typing import Dict

from xrpc.dsl import rpc, RPCType, regular, signal
from xrpc.error import TerminationException, HorizonPassedError
from xrpc.runtime import sender, service

# todo: please note that Request-Reply pattern does would not work with a service that tries
# todo: to access itself.

# todo: the issue is actually that not only the request-reply pattern wouldn't work
# todo: but also the fact that an RPC might have circular dependencies
from xrpc.transport import Origin


class ExemplaryRPC:
    def __init__(self):
        # todo save the required local state here
        self.should_exit = False

    @rpc(RPCType.Signalling)
    def move_something(self, id: int, *xyzargs: int, pop: str, pip: int = 2, **zzargs: int):
        #print('call made', 'ms', id, xyzargs, pop, pip, zzargs)
        # so we pack a call with args and kwargs and let the deserializer guess the contents
        # how do we write a proper deserializer in such a scenario?
        pass

        # todo: we need an ability to save the sender
        # todo: we need an ability to automatically transform the sender to a relevant object

    @rpc(RPCType.Repliable)
    def reply(self, id: int, *xyzargs: int, pop: str, pip: int = 2, **zzargs: int) -> float:
        # so we pack a call with args and kwargs and let the deserializer guess the contents
        # how do we write a proper deserializer in such a scenario?

        return random.random()
        pass

    @regular()
    def regularly_executable(self, id: int = 1) -> int:
        return 1

    @regular()
    def regularly_executable_def(self, id: int = 1, b=6, a=5) -> int:
        return 2

    @regular()
    def regularly_executable_def2(self, id: int = 1, b=6, a=5) -> float:
        return 3

    @rpc(RPCType.Repliable)
    def exit(self):
        self.should_exit = True

    @regular()
    def exit_checket(self) -> float:
        if self.should_exit:
            raise TerminationException()
        return 1

    @signal()
    def on_exit(self) -> bool:
        # return True if we'd like to actually exit.
        # todo: save the relevant local state here.
        return True


class BroadcastClientRPC:
    def __init__(self, broadcast_addr: Origin):
        self.broadcast_addr = broadcast_addr
        self.pings_remaining = 5

    @rpc(type=RPCType.Signalling)
    def ping(self):
        self.pings_remaining -= 1
        logging.getLogger(__name__ + '.' + self.__class__.__name__).info(f'%d', self.pings_remaining)

        if self.pings_remaining <= 0:
            raise TerminationException()

    @regular()
    def broadcast(self) -> float:
        while True:
            s = service(BroadcastRPC, self.broadcast_addr)
            s.arrived()

            return 0.05


class BroadcastRPC:
    def __init__(self):
        self.origins: Dict[Origin, int] = {}
        self.origins_met = set()

    @rpc(type=RPCType.Signalling)
    def arrived(self):
        sdr = sender()
        self.origins[sdr] = 5

        if sdr not in self.origins_met:
            self.origins_met.add(sdr)

    @regular()
    def broadcast(self) -> float:
        for x in self.origins:
            c = service(BroadcastClientRPC, x)
            c.ping()

        for k in list(self.origins.keys()):
            self.origins[k] -= 1

            if self.origins[k] <= 0:
                del self.origins[k]

        logging.getLogger(__name__ + '.' + self.__class__.__name__).info(
            f'%d %s %s',
            len(self.origins_met), self.origins_met, self.origins)

        if len(self.origins_met) == 1 and len(self.origins) == 0:
            raise TerminationException()

        return 0.05



import signal as signalz
from enum import Enum
from typing import NamedTuple, Tuple


class RPCType(Enum):
    """The calling convention of the RPC point"""

    Repliable = 1
    """Reply is expected from the receiver (OK-RECEIVED)
    Beware this does dead-lock services when they both try to send a repliable request at the same time to each other
    """
    Durable = 2
    """we only make sure the packet is received and do not wait for reply (UNDECIDED-RECEIVED)"""
    Signalling = 3
    """we don't care if the packet is received (UNDECIDED-UNDECIDED)"""

    def __repr__(self):
        return str(self.name)


ATTR_RPC = 'rpc_conf'
ATTR_REGULAR = 'rpc_regular'
ATTR_SIGNAL = 'rpc_signal'
ATTR_STARTUP = 'rpc_startup'
ATTR_SOCKETIO = 'rpc_socketio'


class rpc(NamedTuple):
    type: RPCType = RPCType.Repliable
    """Selected calling convention for the RPC call"""

    # add 'RPC GROUPS' to allow certain ports to be used differently

    # todo: rpcs returning iterables

    def __call__(self, fn):
        setattr(fn, ATTR_RPC, self)
        return fn


class regular(NamedTuple):
    initial: float = 0.
    """Initial wait time in seconds"""
    tick: bool = False
    """Run this function on every tick (this should affect the wait times)"""

    def __call__(self, fn):
        setattr(fn, ATTR_REGULAR, self)
        return fn


class signal(NamedTuple):
    codes: Tuple[int] = (signalz.SIGTERM, signalz.SIGINT)
    """Connect to this signal number"""

    def __call__(self, fn):
        setattr(fn, ATTR_SIGNAL, self)
        return fn


class startup(NamedTuple):
    empty: int = 0

    def __call__(self, fn):
        setattr(fn, ATTR_STARTUP, self)
        return fn


# wait for a decision on file descriptor.
class socketio():
    empty: int = 0

    def __call__(self, fn):
        setattr(fn, ATTR_SOCKETIO, self)
        return fn

# todo unhandled_exception handler
# todo startup handler
# todo exit handler



import signal
from contextlib import contextmanager
from datetime import datetime
from typing import Union, Callable, Any

import pytz


@contextmanager
def signal_context(signals=(signal.SIGINT, signal.SIGTERM), handler: Union[int, Any] = signal.SIG_IGN):
    prev_hdlrs = [signal.signal(s, handler) for s in signals]

    try:
        yield prev_hdlrs
    finally:
        for s, prev_hdlr in zip(signals, prev_hdlrs):
            signal.signal(s, prev_hdlr)


def time_now() -> datetime:
    return datetime.now(tz=pytz.utc)


def time_parse(v, format) -> datetime:
    return pytz.utc.localize(datetime.strptime(v, format))



import base64
import logging
import os
import pickle
import zlib
from contextlib import ExitStack, contextmanager

import signal

import subprocess
import sys
from typing import NamedTuple, Callable, Any, Tuple, Dict, List

from xrpc.util import signal_context


@contextmanager
def cov():
    import coverage

    cov = None
    try:
        cov = coverage.process_startup()
        yield
    finally:
        if cov:
            cov.save()


def argv_encode(x: Any):
    return base64.b64encode(zlib.compress(pickle.dumps(x), level=9)).decode()


def argv_decode(x: Any):
    return pickle.loads(zlib.decompress(base64.b64decode(x)))


class PopenStruct(NamedTuple):
    fn: Callable
    args: Tuple[Any]
    kwargs: Dict[str, Any]


def popen(fn, *args, **kwargs) -> subprocess.Popen:
    args = [
        sys.executable,
        '-m',
        __name__,
        f'{fn.__module__}.{fn.__name__}',
        argv_encode(PopenStruct(
            fn,
            args,
            kwargs
        )),
    ]

    logging.getLogger(__name__).debug('Start %s', args)

    p = subprocess.Popen(args)
    return p


class PopenStack:
    """This is mainly for GC purposes. We wait instead of SIGKILL immediately."""

    def __init__(self, timeout=None):
        self.stack: List[subprocess.Popen] = []
        self.timeout = timeout

    def add(self, x: subprocess.Popen):
        self.stack.append(x)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logging.getLogger(__name__).debug('Exit %s %s %s', exc_type, exc_val, exc_tb)
        if exc_type:
            for x in self.stack:
                x.kill()
        else:
            try:
                for x in self.stack:
                    x.wait(timeout=self.timeout)
            except subprocess.TimeoutExpired:
                for x in self.stack:
                    x.send_signal(signal.SIGKILL)
                raise ValueError('We have timed out and therefore killed all of the subprocesses')


def popen_main():
    with ExitStack() as es:
        if os.environ.get('COVERAGE_PROCESS_START'):
            es.enter_context(cov())

        def popen_signal_handler(code, frame):
            if callable(prev_handler[code]):
                prev_handler[code](code, frame)

        codes = (signal.SIGINT, signal.SIGTERM)

        with signal_context(signals=codes, handler=popen_signal_handler) as prev_handlers:
            prev_handler = {k: v for k, v in zip(codes, prev_handlers)}

            try:
                defn: PopenStruct = argv_decode(sys.argv[2])
            except:
                import traceback
                import pprint

                fmtd = pprint.pformat(sys.argv)
                traceback.print_exc(file=sys.stderr)
                sys.stderr.flush()
                raise ValueError(f'Cannot unpickle arguments, called with {fmtd}')

            fn = defn.fn

            fn(*defn.args, **defn.kwargs)


if __name__ == '__main__':
    popen_main()



from xrpc.serde import types
from xrpc.serde.abstract import SerdeInst

SERVER_SERDE_INST = SerdeInst([
    types.NoneSerde(),
    types.TypeVarSerde(),
    types.CallableArgsSerde(),
    types.CallableRetSerde(),
    types.ForwardRefSerde(),
    types.OptionalSerde(),
    types.UnionSerde(),
    types.BytesSerde(),
    types.DateTimeSerde(),
    types.AtomSerde(),
    types.UUIDSerde(),
    types.ListSerde(),
    types.DictSerde(),
    types.EnumSerde(),
    types.DataclassSerde(),
    types.NamedTupleSerde(),
    types.TupleSerde(),
])



import logging
import sys
from contextlib import contextmanager
from urllib.parse import urlparse, ParseResult, parse_qs

import argparse
import threading
from dataclasses import dataclass
from pygelf import GelfUdpHandler
from typing import NamedTuple, List, Union, Optional, Dict, Callable, TypeVar, Tuple


class LoggerLevel(NamedTuple):
    name: Optional[str]
    level: int

    @property
    def level_name(self) -> str:
        return logging._levelToName.get(self.level, str(self.level))

    def __repr__(self):
        prefix = ''
        if self.name:
            prefix = self.name + '='
        return prefix + self.level_name


LoggerLevelList = List[LoggerLevel]


class LoggerSetup(NamedTuple):
    level: LoggerLevel
    level_names: LoggerLevelList
    urls: List[str]

    @classmethod
    def current(cls):
        return LoggerSetup(
            LL(None, logging.getLogger().level),
            [LL(name, logger.level) for name, logger in logging.root.manager.loggerDict.items()],
            []
        )

    def set_level(self, logger: logging.Logger) -> LoggerLevel:
        r = logger.level
        logger.setLevel(self.level.level)
        return LoggerLevel(None, r)

    def set_level_names(self, logger: logging.Logger, should_log=True) -> List[LoggerLevel]:
        lls: Dict[str, LoggerLevel] = {v.name: v for v in self.level_names}

        if should_log:
            logging.getLogger(__name__).debug('Root logger set at level %s', self.level.level_name)

        r = []
        for logger_name in sorted(lls.keys()):
            ll = lls[logger_name]

            if should_log:
                logging.getLogger(__name__).debug('Logger `%s` set at level %s', logger_name, ll.level_name)

            r.append(LoggerLevel(logger_name, logging.getLogger(logger_name).level))
            logger.getChild(logger_name).setLevel(ll.level)
        return r

    def set_handlers(self, logger: logging.Logger) -> List[logging.Handler]:
        r = []

        log_urls = self.urls

        if not len(log_urls):
            log_urls = ['stream:///stderr']

        for url in log_urls:
            o: ParseResult = urlparse(url)
            r.append(LOGGERS[o.scheme](logger, url))
        return r


def LL(name, level: Union[int, str]) -> LoggerLevel:
    if isinstance(level, str):
        level = logging._nameToLevel[level]

    return LoggerLevel(name, level)


DEFAULT_LOGGER_DEFN = [
    # paramiko shits in INFO all the time
    LL('paramiko', logging.ERROR),
]


def type_level(level, name=None) -> LoggerLevel:
    level = level.upper()

    try:
        return LoggerLevel(name, logging._nameToLevel[level])
    except KeyError:
        raise argparse.ArgumentTypeError(level)


def type_level_name(defn) -> LL:
    logger_name, logger_level = defn.split('=', maxsplit=2)

    return type_level(logger_level, name=logger_name)


def type_logger(url) -> str:
    o: ParseResult = urlparse(url)

    if o.scheme in LOGGERS:
        return url
    else:
        raise argparse.ArgumentTypeError(url.scheme)


LOGGER_PREFIX = 'logger_'


def logging_parser(parser: argparse.ArgumentParser, preset_levels: LoggerLevelList = DEFAULT_LOGGER_DEFN):
    for x in preset_levels:
        assert isinstance(x, LoggerLevel), x

    if preset_levels is None:
        preset_levels = []

    parser = parser.add_argument_group('logging', 'Logging setup group')

    parser.add_argument(
        '-L',
        '--log',
        dest=f'{LOGGER_PREFIX}level',
        type=type_level,
        default=LoggerLevel(None, logging.INFO),
        metavar="LEVEL",
        help="Python logging level (default:  %(default)s)",
        choices=[LL(None, x) for x in logging._levelToName.values()]
    )

    parser.add_argument(
        '-LL',
        dest=f'{LOGGER_PREFIX}level_names',
        type=type_level_name,
        default=preset_levels,
        metavar="LEVEL_SPEC",
        help="Python logging levels per logger name  as per `paramiko.transport=DEBUG` (default:  %(default)s)",
        action='append',
    )

    # how do we manage something like this ?

    parser.add_argument(
        '--logger',
        dest=f'{LOGGER_PREFIX}urls',
        default=[],
        action="append",
        type=type_logger,
        help="Should log to stdout. Not setting any of the loggers overrides this to True",
    )


LoggerHandlerFn = Callable[[logging.Logger, str], logging.Handler]


def _logger_stream(logger: logging.Logger, url: str) -> logging.Handler:
    o: ParseResult = urlparse(url)

    AVAIL = {
        'stderr': sys.stderr,
        'stdout': sys.stdout,
    }

    ch = logging.StreamHandler(AVAIL[o.path.lstrip('/')])

    format = logging.Formatter("[%(asctime)s][%(levelname)s][%(name)s]\t%(message)s")
    ch.setFormatter(format)
    ch.setLevel(logging.NOTSET)

    logger.addHandler(ch)

    return ch


def _logger_gelf(logger: logging.Logger, url: str) -> logging.Handler:
    def _split_path_fields(path):
        def by_len(x):
            if len(x) == 2:
                return x
            elif len(x) == 1:
                return [x[0], "1"]
            else:
                return [None, None]

        return {
            k: v for k, v in
            (by_len(x) for x in (x.split('=', 1) for x in path.split('/') if len(x)))
            if k
        }

    o: ParseResult = urlparse(url)

    kwds_a = _split_path_fields(o.path)

    kwds_b = {}

    if o.query:
        kwds_b = {k: v[-1] if v else '1' for k, v in parse_qs(o.query, keep_blank_values=True).items()}

    kwds = {**kwds_a, **kwds_b}

    ch = GelfUdpHandler(host=o.hostname, port=o.port, chunk_size=1350, debug=True,
                        static_fields=kwds)

    logger.addHandler(ch)

    return ch


LOGGERS: Dict[str, LoggerHandlerFn] = {
    'gelf+udp': _logger_gelf,
    'stream': _logger_stream,
}

TL = threading.local()
TL.logging: Optional[LoggerSetup] = None


def logging_config():
    if TL.logging:
        return TL.logging
    else:
        logging.warning('logging_config called without previously setting up the logging')
        return LoggerSetup.current()


@contextmanager
def logging_setup(conf: LoggerSetup, root: Optional[logging.Logger] = None):
    if root is None:
        root = logging.getLogger()

    a = conf.set_level(root)

    try:
        handlers = conf.set_handlers(root)
        try:
            c = conf.set_level_names(root)
            try:
                last_conf = TL.logging
                TL.logging = conf
                try:
                    yield root
                finally:
                    TL.logging = last_conf
            finally:
                LoggerSetup(a, c, []).set_level_names(root, should_log=False)
        finally:
            for handler in handlers:
                root.removeHandler(handler)
    finally:
        logging.getLogger().setLevel(a.level)


@dataclass
class logging_spec:
    setup: LoggerSetup

    def __call__(self, fn):
        from functools import wraps

        @wraps(fn)
        def wrapped(*args, **kwargs):
            with logging_setup(self.setup) as x:
                return fn(*args, **kwargs)

        return wrapped


T = TypeVar('T')


def _dict_split_prefix(dict: Dict[str, T], prefix: str) -> Tuple[Dict[str, T], Dict[str, T]]:
    a = {}
    b = {}

    for k, v in dict.items():
        if k.startswith(prefix):
            a[k[len(prefix):]] = v
        else:
            b[k] = v
    return a, b


@contextmanager
def circuitbreaker(main_logger='main'):
    try:
        yield
    except SystemExit as e:
        # do not spam the console if ``exit`` is called
        exit(e.code)
    except:
        logging.getLogger(main_logger).exception('Main routine caused an exception')
        exit(-1)


def cli_main(fn, parser: argparse.ArgumentParser, argv=None):
    import sys

    if argv is None:
        argv = sys.argv[1:]

    args: argparse.Namespace = parser.parse_args(argv)

    logger_args, fn_args = _dict_split_prefix(vars(args), f'{LOGGER_PREFIX}')

    fn_args[LOGGER_PREFIX] = LoggerSetup(**logger_args)

    with logging_setup(fn_args[LOGGER_PREFIX]):
        with circuitbreaker():
            fn(**fn_args)



import contextlib
import logging
import signal
from collections import deque
from datetime import timedelta, datetime
from functools import partial
from typing import List, Dict, Optional, Callable, Tuple, Any, Deque, NamedTuple

from xrpc.util import signal_context, time_now
from xrpc.dict import RPCLogDict, ObjectDict
from xrpc.error import HorizonPassedError, InvalidFingerprintError, InternalError, TerminationException
from xrpc.net import RPCPacket, RPCReply, RPCPacketType
from xrpc.dsl import RPCType
from xrpc.runtime import ExecutionContext
from xrpc.service import ServiceDefn
from xrpc.transform import get_regular, get_rpc, get_signal, get_startup, get_socketio
from xrpc.transport import Transport, RPCTransportStack, RPCPacketRaw, select_helper


def check_callable_args(fn: Callable, args, kwargs):
    # we only check their existence, not their types
    pass


PrevSignals = Dict[Tuple[str, int], Callable]
ExecutionContextCallable = Callable[[], ExecutionContext]


class SpecialException(Exception):
    pass


def signal_handler_wrapper(code: int, frame, key: str, conf, fn, fn_ec: ExecutionContextCallable,
                           prev_signals: PrevSignals, running_instance=None):
    should_pass = fn_ec().exec('sig', fn, running_instance)
    logging.getLogger('sig.wrapper').debug('Code=%s Frame=%s ShouldPass=%s', code, frame, should_pass)
    if should_pass:
        prev_signals[(key, code)](code, frame)


def signal_handler_default(code: int, frame, state: ObjectDict, prev_signals: PrevSignals):
    logging.getLogger('sig.default').warning('Code=%s Frame=%s', code, frame)

    if code in prev_signals:
        fn = prev_signals[code]
        if callable(fn):
            logging.getLogger('sig.default.prev').warning('Run previous handler %s', fn)
            fn(code, frame)

    state.is_running = False
    raise SpecialException()


@contextlib.contextmanager
def special_handler():
    try:
        yield
    except (KeyboardInterrupt, SpecialException):
        raise KeyboardInterrupt from None


def run_server(cls, running_instance, bind_urls: List[str], horizon_each=60.):
    # todo: enable mapping between RPC groups and the bind urls
    # todo: - the issue with this is the back-comm. ? which bound url sends the packets

    regulars = get_regular(cls)
    rpcs = get_rpc(cls)
    signals = get_signal(cls)
    startups = get_startup(cls)
    socketios = get_socketio(cls)

    # we can't yet build an RPC from the class without pushing the envelope.

    service_defn = ServiceDefn.from_obj(cls)

    waiting_for_regulars: Dict[str, datetime] = {
        k: time_now() + timedelta(seconds=x.conf.initial) for k, x in regulars.items()
    }

    log_dict = RPCLogDict(time_now())

    state = ObjectDict(is_running=True)

    with contextlib.ExitStack() as stack:
        prev_signals2: PrevSignals = {}

        codes = (signal.SIGTERM, signal.SIGINT)

        prev_hdlrs = stack.enter_context(
            signal_context(
                signals=codes,
                handler=partial(signal_handler_default, state=state, prev_signals=prev_signals2)
            )
        )

        for code, prev_hdlr in zip(codes, prev_hdlrs):
            prev_signals2[code] = prev_hdlr
            logging.getLogger('signal.bind.default').debug('Code=%d', code)


        # build transports

        transports = []
        for url, t in ((url, Transport.from_url(url)) for url in bind_urls):
            stack.enter_context(t)

            logging.getLogger('transport.bind').debug('%s', url)
            transports.append(t)

        assert len(transports), 'No transports specified'

        # build signals

        prev_signals: PrevSignals = {}

        def exec_ctx_fn():
            return ExecutionContext(transport_stack, None)

        for k, (conf, fn) in signals.items():
            x = signal_context(
                signals=conf.codes,
                handler=partial(
                    signal_handler_wrapper, key=k, conf=conf, fn=fn,
                    prev_signals=prev_signals,
                    fn_ec=exec_ctx_fn,
                    running_instance=running_instance
                )
            )
            prev_hdlrs = stack.enter_context(x)
            # x is now prev signal handler

            for code, prev_hdlr in zip(conf.codes, prev_hdlrs):
                prev_signals[(k, code)] = prev_hdlr
                logging.getLogger('signal.bind').debug('Code=%d Name=%s', code, k)

        # build transport stack

        class ExecutionQueueEntry(NamedTuple):
            p: RPCPacket
            fb: Any
            raw_packet: RPCPacketRaw
            args: Any
            kwargs: Any

        execution_queue: Deque[ExecutionQueueEntry] = deque()

        def recv_packet(raw_packet: RPCPacketRaw):
            p = raw_packet.packet

            rp: Optional[RPCPacket] = None

            try:
                if p.key in log_dict:
                    rep = log_dict[p.key]

                    if rpcs[p.name].conf.type != RPCType.Signalling:
                        rp = RPCPacket(p.key, RPCPacketType.Rep, RPCReply.ok.value, rep)
                else:
                    if p.name not in rpcs:
                        logging.error(f'Could not find name %s %s', p.name, set(rpcs.keys()))
                        raise InvalidFingerprintError('name')

                    fa, fb = service_defn.rpcs_serde[p.name]

                    try:
                        args, kwargs = service_defn.serde.deserialize(fa, p.payload)
                    except Exception as e:
                        # could not deserialize the payload correctly
                        logging.exception(f'Failed to deserialize packet from {raw_packet.addr}')
                        raise InvalidFingerprintError('args')

                    # todo we have deserialized the packet

                    execution_queue.append(
                        ExecutionQueueEntry(
                            p,
                            fb, raw_packet, args, kwargs
                        )
                    )

                    if rpcs[p.name].conf.type == RPCType.Durable:
                        rp = RPCPacket(p.key, RPCPacketType.Rep, RPCReply.ok.value, None)
                        log_dict[p.key] = None
                    elif rpcs[p.name].conf.type == RPCType.Signalling:
                        log_dict[p.key] = None
            except InvalidFingerprintError as e:
                rp = RPCPacket(p.key, RPCPacketType.Rep, RPCReply.fingerprint.value,
                               service_defn.serde.serialize(Optional[str], e.reason))
            except HorizonPassedError as e:
                rp = RPCPacket(p.key, RPCPacketType.Rep, RPCReply.horizon.value,
                               service_defn.serde.serialize(datetime, e.when))

            if rp:
                transport_stack.send(RPCPacketRaw(raw_packet.addr, rp))

        transport_stack = RPCTransportStack(transports)

        transport_stack.push(lambda x: x.packet.type == RPCPacketType.Req, recv_packet)

        # build startups

        for k, (conf, fn) in startups.items():
            pass

        # build socketios

        socketio_states = {}

        for k, (conf, fn) in socketios.items():
            socketio_states[k] = fn(running_instance)

        # end of builds

        stack.enter_context(special_handler())

        while state.is_running:
            step_time = time_now()

            max_poll_regulars = {k: (x - step_time).total_seconds() for k, x in waiting_for_regulars.items()}
            max_poll_horizon = ((log_dict.horizon + timedelta(seconds=horizon_each) * 2) - step_time).total_seconds()

            items = list(max_poll_regulars.values()) + [max_poll_horizon]

            max_poll = min(items)

            fds_transport = [x.fd for x in transport_stack.transports]
            fds_socketios = [v for _, v in sorted(socketio_states.items())]

            flags = select_helper(fds_transport + fds_socketios, max_wait=max(0., max_poll))

            flags_transport = flags[:len(fds_transport)]
            flags_socketios = flags[len(fds_transport):]

            should_receive = any(flags_transport)
            should_regular = any(x <= 0 or regulars[k].conf.tick for k, x in max_poll_regulars.items())
            should_horizon = max_poll_horizon <= 0
            should_socketio = any(flags_socketios)

            assert len(transport_stack.stack) == 1, transport_stack.stack

            if should_receive:
                transport_stack.recv()

                # todo process the queue

                while len(execution_queue):
                    eq = execution_queue.popleft()

                    p = eq.p
                    fb = eq.fb
                    raw_packet = eq.raw_packet
                    args = eq.args
                    kwargs = eq.kwargs

                    rp: Optional[RPCPacket] = None

                    try:
                        has_returned = False

                        def replier(ret: Any):
                            assert rpcs[p.name].conf.type == RPCType.Repliable, p.name

                            nonlocal has_returned
                            assert not has_returned

                            ret_payload = service_defn.serde.serialize(fb, ret)

                            log_dict[p.key] = ret_payload

                            rp = RPCPacket(p.key, RPCPacketType.Rep, RPCReply.ok.value, ret_payload)

                            transport_stack.send(RPCPacketRaw(raw_packet.addr, rp))
                            has_returned = True

                        ctx = ExecutionContext(transport_stack, raw_packet.addr, p.key, ret=replier)
                        rpc_fn = rpcs[p.name].fn
                        try:
                            ret = ctx.exec('call', rpc_fn, running_instance, *args, **kwargs)
                        except TerminationException:
                            raise
                        except Exception as e:
                            logging.getLogger(__name__).exception(
                                'While receiving the payload [%s][%s][%s]', p.name, args, kwargs)
                            raise InternalError(str(e))

                        if rpcs[p.name].conf.type == RPCType.Repliable:
                            if not has_returned:
                                replier(ret)
                            elif has_returned and ret is None:
                                pass
                            elif has_returned and ret is not None:
                                raise ValueError()
                            else:
                                raise NotImplementedError()
                    except InternalError as e:
                        rp = RPCPacket(p.key, RPCPacketType.Rep, RPCReply.internal.value, e.reason)
                        raise
                    except TerminationException as e:
                        state.is_running = False
                        raise SpecialException()
                        continue
                    finally:
                        if rp and rpcs[p.name].conf.type == RPCType.Repliable:
                            transport_stack.send(RPCPacketRaw(raw_packet.addr, rp))

            if should_regular:
                for name, regular, callable in [(k, *regulars[k]) for k, v in max_poll_regulars.items() if v <= 0.]:
                    try:
                        x: float = ExecutionContext(transport_stack).exec('reg', callable, running_instance)
                    except TerminationException:
                        state.is_running = False
                        raise SpecialException()
                    else:
                        waiting_for_regulars[name] = time_now() + timedelta(seconds=x)

            if should_horizon:
                log_dict.set_horizon(step_time - timedelta(seconds=horizon_each))

            if should_socketio:
                for is_ready, key in zip(flags_socketios, [k for k, _ in sorted(socketio_states.items())]):
                    if is_ready:
                        callable = socketios[key][1]

                        r = ExecutionContext(transport_stack).exec('io', callable, running_instance)

                        assert r is not None, key

                        socketio_states[key] = r



import logging
import sys
from collections import deque
from types import ModuleType

from xrpc.serde.error import SerdeException

if sys.version_info >= (3, 7):
    from typing import NamedTuple, Any, List, ForwardRef, _type_check, Optional, Union, Dict, Callable, \
        _tp_cache, TypeVar, Type, Tuple, Deque, Set
else:
    from typing import Any, _ForwardRef, _FinalTypingBase, Optional, Union, List, Dict, _tp_cache, _type_check, \
        Callable, \
        NamedTuple, Tuple, TypeVar, Type

from dataclasses import dataclass, field

DESER = Callable[[Union[list, dict]], Any]
SER = Callable[[Any], Union[list, dict]]


def _repr_typevar(x):
    if isinstance(x, TypeVar):
        return f'TypeVar({repr(x)}, {id(x)})'
    else:
        return repr(x)


@dataclass
class SerdeStepContext:
    t: Any = None
    mod: ModuleType = None
    generic_vals: Dict[Any, Any] = field(default_factory=dict)

    def __repr__(self):

        def map_repr(x):
            if isinstance(x, dict):
                items = ['='.join((_repr_typevar(k), _repr_typevar(v))) for k, v in x.items()]
                items = ', '.join(items)
                return f'{{{items}}}'
            else:
                return repr(x)

        items = [self.t, self.mod, self.generic_vals]
        items = [map_repr(x) for x in items if x is not None]
        items = ', '.join(items)
        return f'{self.__class__.__qualname__}({items})'

    def merge(self, other: 'SerdeStepContext'):
        return SerdeStepContext(other.t, other.mod, {**self.generic_vals, **other.generic_vals})


@dataclass
class SerdeNode:
    type: Any
    deps: List[DESER]
    ctx: Optional[SerdeStepContext] = field(default_factory=SerdeStepContext)


class SerdeInst:
    def __init__(self, context: List['SerdeType']):
        self.context = context

    def match(self, t: Any, ctx: SerdeStepContext) -> 'SerdeType':
        for x in self.context:
            try:
                if x.match(t, ctx):
                    logging.getLogger(__name__ + '.match').debug('%s -> %s %s', t, x, ctx)
                    return x
            except:
                raise ValueError(f'GivenClass={t.__class__} GivenType={t} Matcher={x}')
        else:
            raise ValueError(f'Could not match `{t}` `{t.__class__}`')

    def norm(self, t: Any, ctx: SerdeStepContext) -> Any:
        r = self.match(t, ctx).norm(self, t, ctx)
        logging.getLogger(__name__ + '.norm').debug('%s -> %s [%s]', t, r, ctx)
        return r

    def step(self, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        return self.match(t, ctx).step(self, t, ctx)

    def deserializer(self, t: Any, deps: List[DESER]) -> DESER:
        return self.match(t, SerdeStepContext()).deserializer(t, deps)

    def serializer(self, t: Any, deps: List[SER]) -> SER:
        return self.match(t, SerdeStepContext()).serializer(t, deps)


class SerdeTypeDeserializer:
    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.parent}, {str(self.t)}, {[x.__class__.__name__ for x in self.deps]})'

    def __repr__(self) -> str:
        return str(self)

    def __init__(self, parent: 'SerdeType', t: Any, deps: List[DESER]):
        self.parent = parent
        self.t = t
        self.deps = deps

    def __call__(self, val: Union[list, dict]) -> Any:
        raise NotImplementedError(f'Deserializer {self.parent.__class__.__name__}')


class SerdeTypeSerializer:

    def __str__(self) -> str:
        return SerdeTypeDeserializer.__str__(self)

    def __init__(self, parent: 'SerdeType', t: Any, deps: List[SER]):
        self.parent = parent
        self.t = t
        self.deps = deps

    def __call__(self, val: Any) -> Union[list, dict]:
        raise NotImplementedError(f'Serializer {self.parent.__class__.__name__}')


class SerdeType:
    """
    This is an instantiable Deserializer Type
    """
    type: str = None
    cls_serializer: Type[SerdeTypeSerializer] = SerdeTypeSerializer
    cls_deserializer: Type[SerdeTypeDeserializer] = SerdeTypeDeserializer

    def match(self, t: Any, ctx: SerdeStepContext) -> bool:
        pass

    def norm(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> Any:
        return t

    def step(self, i: SerdeInst, t: Any, ctx: SerdeStepContext) -> SerdeNode:
        raise NotImplementedError(self.__class__.__name__)

    # what if the passable object to the serializer
    # is not the same as the object passed to the deserializer ?
    # - it does not matter as we are telling the engine upfront
    def deserializer(self, t: Any, deps: List[DESER]) -> DESER:
        return self.cls_deserializer(self, t, deps)

    def serializer(self, t: Any, deps: List[SER]) -> SER:
        return self.cls_serializer(self, t, deps)


T = TypeVar('T')


class SerdeStruct(NamedTuple):
    deserializers: Dict[Any, DESER]
    serializers: Dict[Any, SER]

    def deserialize(self, t: Type[T], val) -> T:
        try:
            return self.deserializers[t](val)
        except Exception as e:
            raise SerdeException(val, '$', t=t, par=e)

    def serialize(self, t, val):
        try:
            return self.serializers[t](val)
        except Exception as e:
            raise SerdeException(val, '$', t=t, par=e)


@dataclass
class SerdeSet:
    items: List[SerdeNode] = field(default_factory=list)

    def __iter__(self):
        return iter(self.items)

    def merge(self, *xs: 'SerdeSet') -> 'SerdeSet':
        r: List[SerdeNode] = []

        for x in (self,) + xs:
            for y in x:
                for rx in r:
                    if rx.type == y.type:
                        break
                else:
                    r.append(y)

        return SerdeSet(r)

    def struct(self, i: SerdeInst) -> SerdeStruct:
        desers_pre = {}
        sers_pre = {}

        def raiser_deser(val: dict) -> Any:
            assert False, 'should never happen'

        def raiser_ser(val: dict) -> Any:
            assert False, 'should never happen'

        for x in self.items:
            desers_pre[x.type] = [raiser_deser for _ in x.deps]
            sers_pre[x.type] = [raiser_ser for _ in x.deps]

        desers = {}
        sers = {}

        for t, deps in desers_pre.items():
            desers[t] = i.deserializer(t, deps)

        for t, deps in sers_pre.items():
            sers[t] = i.serializer(t, deps)

        for n, (t, deps) in zip(self.items, desers_pre.items()):
            assert len(n.deps) == len(deps), (n.deps, deps)
            for dt, i in zip(n.deps, range(len(deps))):
                deps[i] = desers[dt]

        for n, (t, deps) in zip(self.items, sers_pre.items()):
            assert len(n.deps) == len(deps), (n.deps, deps)
            for dt, i in zip(n.deps, range(len(deps))):
                deps[i] = sers[dt]

        return SerdeStruct(desers, sers)

    @classmethod
    def empty(cls):
        return SerdeSet([])

    @classmethod
    def walk(cls, i: SerdeInst, t: Any, ctx: SerdeStepContext = SerdeStepContext()) -> 'SerdeSet':
        assert isinstance(ctx, SerdeStepContext)

        def do_visit(obj, ctx):
            x = i.step(obj, ctx)

            r[x.type] = x

            for dep in x.deps:
                if dep not in seen:
                    seen.add(dep)
                    to_visit.append((dep, x.ctx))

        seen: Set[Any] = {t, }
        r: Dict[Any, SerdeNode] = {}

        to_visit: Deque[Tuple[t, SerdeStepContext]] = deque()

        do_visit(t, ctx)

        while len(to_visit):
            x, ctx = to_visit.popleft()
            do_visit(x, ctx)

        return SerdeSet(list(r.values()))



import logging
from urllib.parse import urlparse, parse_qs, ParseResult, urlunparse

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, NamedTuple

from xrpc.dsl import RPCType
from xrpc.error import TimeoutError, InvalidFingerprintError, HorizonPassedError, InternalError
from xrpc.net import RPCKey, RPCPacket, RPCReply, RPCPacketType
from xrpc.service import ServiceDefn
from xrpc.transport import Origin, RPCTransportStack, RPCPacketRaw, select_helper
from xrpc.util import time_now


class ClientConfig(NamedTuple):
    timeout_resend: float = 0.033
    timeout_total: Optional[float] = None
    ignore_horizon: bool = False
    """If ```HorizonPassedError``` is passed, restart the action"""


def dest_overrides(url):
    r: ParseResult = urlparse(url)
    qs = parse_qs(r.query) if r.query else {}
    r = r._replace(query=None)
    r: str = urlunparse(r)

    return {k: v[0] for k, v in qs.items()}, r


class ServiceWrapper:
    def __init__(self, defn: ServiceDefn, conf: ClientConfig, transport: RPCTransportStack, dest: Origin):
        self.transport = transport
        self.overrides, self.dest = dest_overrides(dest)
        self.defn = defn
        self.conf = conf

    def _assert(self):
        missing = []
        for x in self.overrides.keys():
            if x not in self.defn.rpcs:
                missing.append(x)
        if len(missing):
            assert False, missing

    def __getattr__(self, item):
        f1 = item in self.defn.rpcs
        alias = self.overrides.get(item)

        if f1 or alias:
            return CallWrapper(self, item, alias)
        else:
            logging.getLogger(__name__).error('%s', self.defn.rpcs)
            raise AttributeError(item)


@dataclass
class RequestWrapper:
    type: ServiceWrapper
    name: str
    alias: Optional[str]
    key: RPCKey = field(default_factory=RPCKey)

    def __call__(self, *args, **kwargs):
        c = self.type.defn.rpcs[self.name]
        i, o = self.type.defn.rpcs_serde[self.name]
        payload = self.type.defn.serde.serialize(i, [args, kwargs])

        packet = RPCPacket(self.key, RPCPacketType.Req, self.alias or self.name, payload)

        # the only difference between a client and a server is NONE.
        # the only issue would be the routing of the required packets to the required instances

        if c.conf.type == RPCType.Signalling:
            self.type.transport.send(RPCPacketRaw(self.type.dest, packet))
        elif c.conf.type in [RPCType.Repliable, RPCType.Durable]:
            time_started = time_now()

            stop = False
            ret = None

            def process_packet(x: RPCPacketRaw):
                packet = x.packet

                assert packet.key == self.key, (packet, self.key)

                nonlocal stop
                nonlocal ret
                stop = True

                if packet.name == RPCReply.ok.value:
                    ret = self.type.defn.serde.deserialize(o, packet.payload)
                elif packet.name == RPCReply.fingerprint.value:
                    raise InvalidFingerprintError(self.type.defn.serde.deserialize(Optional[str], packet.payload))
                elif packet.name == RPCReply.horizon.value:
                    raise HorizonPassedError(self.type.defn.serde.deserialize(datetime, packet.payload))
                elif packet.name == RPCReply.internal.value:
                    raise InternalError(packet.payload)
                else:
                    raise NotImplementedError(packet.name)

            self.type.transport.push(
                lambda x: x.packet.key == self.key and x.packet.type == RPCPacketType.Rep,
                process_packet
            )

            try:
                while not stop:
                    time_step = time_now()

                    dur_passed = time_step - time_started

                    if self.type.conf.timeout_total is not None and dur_passed.total_seconds() > self.type.conf.timeout_total:
                        raise TimeoutError()

                    self.type.transport.send(RPCPacketRaw(self.type.dest, packet))

                    # todo transform transport sends to
                    # todo an ability to buffer the transport outputs

                    flags = select_helper([x.fd for x in self.type.transport.transports],
                                          max_wait=max(0., self.type.conf.timeout_resend))

                    self.type.transport.recv(flags)
                return ret

            finally:
                # todo: we can not be sure that all of the packets here have been read ? or can we?
                self.type.transport.pop()
        else:
            raise NotImplementedError(c.conf.type)


@dataclass()
class CallWrapper:
    type: ServiceWrapper
    name: str
    alias: Optional[str]

    def __call__(self, *args, **kwargs):
        while True:
            try:
                return RequestWrapper(self.type, self.name, self.alias)(*args, **kwargs)
            except HorizonPassedError:
                if self.type.conf.ignore_horizon:
                    continue
                raise


def build_wrapper(pt: ServiceDefn, transport: RPCTransportStack, dest: Origin, conf: ClientConfig = ClientConfig()):
    return ServiceWrapper(pt, conf, transport, dest)



import argparse

from xrpc.logging import logging_parser, cli_main
from xrpc.server import run_server
from xrpc.examples.exemplary_rpc import ExemplaryRPC


def main(bind, **kwargs):
    rpc = ExemplaryRPC()
    run_server(rpc, bind)


def parser():
    parser = argparse.ArgumentParser()

    logging_parser(parser)

    parser.add_argument(
        '-b',
        '--bind',
        dest='bind',
        action='append',
        default=[],
        help='Bind to these addresses'
    )

    return parser


if __name__ == '__main__':
    cli_main(main, parser())



import logging
import sys
from inspect import getfullargspec
from typing import Dict, Any, Tuple, Callable, NamedTuple

from xrpc.dsl import ATTR_RPC, ATTR_REGULAR, rpc, regular, signal, ATTR_SIGNAL, ATTR_STARTUP, startup, \
    socketio, ATTR_SOCKETIO
from xrpc.generic import build_generic_context


def build_rpc_list(x, conf_attr_name=ATTR_RPC) -> Tuple[str, Any, Any]:
    x, _ = build_generic_context(x)

    if sys.version_info >= (3, 7):
        if hasattr(x, '__origin__'):
            x = x.__origin__

    r = []

    logging.getLogger(f'{__name__}.build_rpc_list').debug('%s %s', x, dir(x))

    for attr_name in dir(x):
        attr_val = getattr(x, attr_name)

        if callable(attr_val) and hasattr(attr_val, conf_attr_name):
            r.append((attr_name, attr_val, getattr(attr_val, conf_attr_name)))

    return r


class RegularConf(NamedTuple):
    conf: regular
    fn: Callable


def get_regular(x) -> Dict[str, RegularConf]:
    rs = build_rpc_list(x, ATTR_REGULAR)
    r = {}

    for n, attr, x in rs:
        argspec = getfullargspec(attr)
        rtn_type = argspec.annotations.get('return', None)

        assert rtn_type in [float, int], f'Incorrect return type for {n}: {rtn_type}'
        r[n] = RegularConf(x, attr)

    return r


def get_signal(x) -> Dict[str, Tuple[signal, Callable]]:
    rs = build_rpc_list(x, ATTR_SIGNAL)
    r = {}

    for n, attr, x in rs:
        argspec = getfullargspec(attr)
        rtn_type = argspec.annotations.get('return', None)

        assert rtn_type in [bool, None], f'Incorrect return type for {n}: {rtn_type}'
        r[n] = (x, attr)

    return r


def get_startup(x) -> Dict[str, Tuple[startup, Callable]]:
    rs = build_rpc_list(x, ATTR_STARTUP)
    r = {}

    for n, attr, x in rs:
        r[n] = (x, attr)

    return r


def get_socketio(x) -> Dict[str, Tuple[socketio, Callable]]:
    rs = build_rpc_list(x, ATTR_SOCKETIO)
    r = {}

    for n, attr, x in rs:
        r[n] = (x, attr)

    return r


class RPCCallable(NamedTuple):
    fn: Callable
    conf: rpc


RPCS = Dict[str, RPCCallable]


def get_rpc(x) -> RPCS:
    rs = build_rpc_list(x)

    return {n: RPCCallable(fn, c) for n, fn, c in rs}



import inspect
import sys
from argparse import ArgumentParser
from typing import Any, TypeVar, Type, Dict, List, Optional

from dataclasses import fields, dataclass, MISSING, _MISSING_TYPE

from xrpc.generic import build_generic_context
from xrpc.logging import _dict_split_prefix
from xrpc.serde.types import is_union

T = TypeVar('T')


@dataclass
class ParsableConf:
    names: List[str]
    type: Optional[Any] = None
    action: Optional[str] = None


def is_list(t):
    t, _ = build_generic_context(t)

    if sys.version_info >= (3, 7):
        if hasattr(t, '__origin__'):
            return t.__origin__ is list
    if inspect.isclass(t):
        return issubclass(t, List)
    else:
        return False


def _guess_type_action(dest, type_, default) -> ParsableConf:
    action = None

    if is_union(type_):
        args = type_.__args__

        if len(args) == 2 and args[-1] == type(None):
            type_ = args[0]
        else:
            assert False, type_

    if type_ is bool:
        type_ = None

        if default is True:
            action = 'store_false'
        else:
            action = 'store_true'
    elif is_list(type_):
        type_, = type_.__args__
        action = 'append'
    return ParsableConf([dest], type_, action)


class Parsable:
    @classmethod
    def overrides(cls) -> Dict[str, ParsableConf]:
        return {}

    @classmethod
    def parser(cls, prefix: str, argparse: ArgumentParser):
        overrides = cls.overrides()

        for f in fields(cls):
            if f.default_factory is not MISSING:
                default = f.default_factory()
            else:
                default = f.default

                assert default is not MISSING, f

            type_ = f.type

            name = f'{f.name}'

            if prefix:
                dest = f'{prefix}_{name}'
            else:
                dest = name

            if name in overrides:
                conf = overrides[name]
            else:
                conf = _guess_type_action('--' + dest, type_, default)

            help = None

            if hasattr(f, '__doc__'):
                help = f.__doc__

            type_action = {}
            if conf.type:
                type_action['type'] = conf.type

            if conf.action:
                type_action['action'] = conf.action

            argparse.add_argument(
                *conf.names,
                dest=dest,
                default=default,
                help=help,
                **type_action
            )

    @classmethod
    def from_parser(cls: Type[T], prefix, d, forget_other=True) -> T:
        filtered_fields, unfiltered_fields = _dict_split_prefix(d, prefix + '_' if prefix else '')

        field_set = {x.name for x in fields(cls)}

        filtered_fields = {k: v for k, v in filtered_fields.items() if k in field_set}
        unfiltered_fields = {k: v for k, v in filtered_fields.items() if k not in field_set}
        unfiltered_fields = {**unfiltered_fields, **unfiltered_fields}

        if forget_other:
            return cls(**filtered_fields)
        else:
            return unfiltered_fields, cls(**filtered_fields)



import hashlib
import struct
from enum import Enum
from typing import NamedTuple, Tuple, List, Optional

from dataclasses import dataclass


class ChecksumType(Enum):
    sha256 = 0

    def __repr__(self):
        return self.__class__.__name__ + '.' + self.name


CS_PACKER = struct.Struct('!B')


@dataclass
class Checksum:
    type: ChecksumType
    payload: bytes

    @classmethod
    def unpack(cls, payload: bytes) -> Tuple['Checksum', bytes]:
        cti, = CS_PACKER.unpack(payload[:CS_PACKER.size])
        t = ChecksumType(cti)

        if t == ChecksumType.sha256:
            ln = 256 // 8
            return Checksum(t, payload[CS_PACKER.size:CS_PACKER.size + ln]), payload[CS_PACKER.size + ln:]
        else:  # pragma: no cover
            assert False, t

    def pack(self) -> bytes:
        return CS_PACKER.pack(self.type.value) + self.payload

    @classmethod
    def new(cls, type: ChecksumType, payload: bytes) -> 'Checksum':
        if type == ChecksumType.sha256:
            m = hashlib.sha256()
            m.update(payload)
            return Checksum(type, m.digest())
        else:  # pragma: no cover
            assert False, type


class BlockError(Exception):
    pass


class ChecksumError(BlockError):
    def __init__(self, curr: Optional[Checksum], reqd: Checksum):
        self.curr = curr
        self.reqd = reqd

        super().__init__()

    def __str__(self):
        return f'Curr={self.curr} Required={self.reqd}'


# todo: we must be able to chain blocks.

BLK_PACKER = struct.Struct('!L')


@dataclass
class Block:
    checksum: Checksum
    payload: bytes

    @classmethod
    def new(cls, payload: bytes, checksum_type: ChecksumType = ChecksumType.sha256):
        return Block(Checksum.new(checksum_type, payload), payload)

    @classmethod
    def unpack(cls, payload: bytes) -> 'Block':
        checksum_reqd, payload = Checksum.unpack(payload)

        payload_size, = BLK_PACKER.unpack(payload[:BLK_PACKER.size])

        if len(payload) < payload_size + BLK_PACKER.size:
            raise ChecksumError(None, checksum_reqd)

        payload_bytes = payload[BLK_PACKER.size:BLK_PACKER.size + payload_size]

        checksum_curr = Checksum.new(checksum_reqd.type, payload_bytes)

        if checksum_reqd != checksum_curr:
            raise ChecksumError(checksum_curr, checksum_reqd)

        return Block(checksum_reqd, payload_bytes)

    def pack(self) -> bytes:
        cs_bytes = self.checksum.pack()

        pl_size = BLK_PACKER.pack(len(self.payload))

        return cs_bytes + pl_size + self.payload



# possible exceptions caused by the API calls
from datetime import datetime
from typing import Optional

from xrpc.util import time_now


class RPCError(Exception):
    pass


class UndecidedError(RPCError):
    # We have no idea if the operation succeedeed or not
    # Operation state: UNDECIDED
    pass


class FailedError(RPCError):
    # Operation FAILED
    pass


class HorizonPassedError(UndecidedError):
    def __init__(self, when: datetime):
        self.when = when
        super().__init__(when)

    def __str__(self):
        now = time_now()

        total = self.when - now

        return f'{self.__class__.__name__}(Seconds={total.total_seconds():0.2f}s)'

    # the API server had already purged everything later than this specific TIME horizon.

    # Please note this does not provide any guarantees of the operation completion
    pass


class TimeoutError(UndecidedError):
    # if the configuration of the callable allows for terminating timeouts,
    # then this is the exception that is caused by them

    # Please note this does not provide any guarantees of the operation completion
    pass


class InternalError(FailedError):
    # a callable raised an exception
    def __init__(self, reason):
        self.reason = reason
        super().__init__(reason)


class InvalidFingerprintError(FailedError):
    # the call was made to a point that does not support given arguments
    def __init__(self, reason: Optional[str] = None):
        self.reason = reason
        super().__init__(reason)


class TerminationException(Exception):
    pass



import argparse
from multiprocessing import Pool

from xrpc.logging import logging_parser, cli_main
from xrpc.server import run_server
from xrpc.examples.exemplary_rpc import BroadcastClientRPC, BroadcastRPC


def run_server_a(b_addr):
    run_server(BroadcastClientRPC(b_addr), ['udp://127.0.0.1:7483'])


def run_server_b(c_addr):
    run_server(BroadcastRPC(), ['udp://127.0.0.1:7482'])


def main(bind, **kwargs):
    with Pool(2) as p:
        a = p.apply_async(run_server_a, args=(('127.0.0.1', 7482),))
        b = p.apply_async(run_server_b, args=(('127.0.0.1', 7482),))

        a.get()
        b.get()


def parser():
    parser = argparse.ArgumentParser()

    logging_parser(parser)

    parser.add_argument(
        '-b',
        '--bind',
        dest='bind',
        action='append',
        default=[],
        help='Bind to these addresses'
    )

    return parser


if __name__ == '__main__':
    cli_main(main, parser())



import logging
import sys
from typing import TypeVar, Tuple, List, Dict
from dataclasses import dataclass, is_dataclass

from xrpc.serde.abstract import SerdeStepContext, _repr_typevar


def build_generic_context(t, ctx=SerdeStepContext()):
    logging.getLogger(__name__ + '.ep').debug('%s %s', t, ctx)

    def mmaps(pars, args):
        maps = dict(zip(pars, args))

        maps = {k: ctx.generic_vals.get(v, v) if isinstance(v, TypeVar) else v for k, v in maps.items()}

        uninst = [k for k in maps if isinstance(maps[k], TypeVar)]

        if len(uninst):
            raise ValueError(f'Not all generic parameters are instantiated: {uninst}, {t} {ctx}')

        return maps

    if sys.version_info >= (3, 7):
        if not hasattr(t, '__origin__'):
            return t, ctx

        if t.__origin__ is tuple:
            params, args = t.__parameters__, t.__args__

            args = params[len(args):] + args

            args = tuple(ctx.generic_vals.get(x, x) for x in args)
            logging.getLogger(__name__ + '.tuple.a').debug('%s', args)
            maps = mmaps(params[len(args):], args)
            t = Tuple[args]
        elif t.__origin__ is list:
            params, args = t.__parameters__, t.__args__

            args = params[len(args):] + args

            args = tuple(ctx.generic_vals.get(x, x) for x in args)
            logging.getLogger(__name__ + '.list.a').debug('%s', args)
            maps = mmaps(params[len(args):], args)
            t = List[args]
        elif is_dataclass(t.__origin__):
            params, args = t.__origin__.__parameters__, t.__args__

            logging.getLogger(__name__ + '.dc.a').debug('%s %s %s', params, args, ctx.generic_vals)

            args = params[len(args):] + args
            args = tuple(ctx.generic_vals.get(x, x) for x in args)

            logging.getLogger(__name__ + '.dc.b').debug('%s %s %s %s', t.__origin__, args, t.__origin__[args], ctx.generic_vals)
            t = t.__origin__[args]
            maps = mmaps(t.__origin__.__parameters__, args)
        elif t.__origin__ is dict:
            args = tuple(ctx.generic_vals.get(x, x) for x in t.__args__)

            t = Dict[args]
            maps = mmaps(t.__parameters__, t.__args__)
        else:
            maps = mmaps(t.__origin__.__parameters__, t.__args__)

        ctx = SerdeStepContext(mod=ctx.mod, generic_vals={**ctx.generic_vals, **maps})

        logging.getLogger(__name__ + '.exit.3.7').debug('%s %s', t, ctx)

        return t, ctx
    else:
        if not hasattr(t, '_gorg'):
            return t, ctx

        pars, args = t._gorg.__parameters__, t.__args__ if t.__args__ else t._gorg.__parameters__

        logging.getLogger(__name__).debug('%s %s %s %s', t, t._gorg.__parameters__, t.__args__, t._gorg.__parameters__)
        logging.getLogger(__name__).debug('%s %s', pars, args)

        maps = mmaps(pars, args)

        if issubclass(t, List):
            t = List[tuple(ctx.generic_vals.get(x, x) for x in t.__args__)]
        elif issubclass(t, Tuple):
            t = Tuple[tuple(ctx.generic_vals.get(x, x) for x in t.__args__)]
        elif is_dataclass(t):
            t = t._gorg[tuple(ctx.generic_vals.get(x, x) for x in t.__args__)]

        return t, SerdeStepContext(mod=ctx.mod, generic_vals={**ctx.generic_vals, **maps})



import base64
import json
import struct
import zlib

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import NamedTuple, Any
from uuid import UUID, uuid4

import bitstruct
import pytz

from xrpc.serde.types import ISO8601
from xrpc.util import time_now

RPC_TS_FORMAT_SPEC = [
    ('year', 12, 12, 1.5, 4096),
    ('month', 4, 16, 2, 16),
    ('day', 5, 21, 2.625, 32),
    ('hour', 5, 26, 3.25, 32),
    ('minute', 6, 32, 4, 64),
    ('second', 6, 38, 4.75, 64),
    ('microsecond', 20, 58, 7.25, 1048576),
    ('_reserved', 6, 64, 8, 64),
]

RPC_TS_FMT = ''.join(f'u{x}' for _, x, *_ in RPC_TS_FORMAT_SPEC)
RPC_TS_FMT_COMPILED = bitstruct.compile(RPC_TS_FMT)
RPC_TS_COMPILED_SIZE = RPC_TS_FMT_COMPILED.calcsize() // 8
RPC_UUID_SIZE = 16
RPC_KEY_SIZE = RPC_TS_COMPILED_SIZE + RPC_UUID_SIZE


def time_pack(dt: datetime):
    timestamp = RPC_TS_FMT_COMPILED.pack(
        *(getattr(dt, x) if not x.startswith('_') else 0 for x, *_ in RPC_TS_FORMAT_SPEC)
    )

    return timestamp


def time_unpack(body: bytes) -> datetime:
    to_unpack = zip(
        (x for x, *_ in RPC_TS_FORMAT_SPEC),
        RPC_TS_FMT_COMPILED.unpack(body[:RPC_TS_COMPILED_SIZE])
    )

    return datetime(**{k: v for k, v in to_unpack if not k.startswith('_')}, tzinfo=pytz.utc)


@dataclass(eq=True, frozen=True)
class RPCKey:
    timestamp: datetime = field(default_factory=time_now)
    uuid: UUID = field(default_factory=uuid4)

    @classmethod
    def new(cls):
        return RPCKey(time_now(), uuid4())

    def pack(self) -> bytes:
        assert self.timestamp.tzinfo == pytz.utc

        timestamp = time_pack(self.timestamp)

        uuid = self.uuid.bytes

        return timestamp + uuid

    def pack_str(self) -> str:
        return base64.b64encode(self.pack(), altchars=b'_-').decode()

    @classmethod
    def unpack(cls, body: bytes):
        timestamp = time_unpack(body[:RPC_TS_COMPILED_SIZE])
        uuid = UUID(bytes=body[RPC_TS_COMPILED_SIZE:])

        return RPCKey(timestamp=timestamp, uuid=uuid)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.timestamp:{ISO8601}} {self.uuid.hex})'


class UUKey(RPCKey):
    def __repr__(self):
        return f'{self.__class__.__name__}({self.pack_str()})'


class RPCReply(Enum):
    ok = 'ok'
    horizon = 'hor'
    fingerprint = 'fin'
    internal = 'int'


RPC_PACKET_PACKER = struct.Struct('!BHL')
RPC_PACKET_FIXED_SIZE = RPC_PACKET_PACKER.size
RPC_PACKET_HEADER_SIZE = RPC_KEY_SIZE + RPC_PACKET_FIXED_SIZE

assert RPC_PACKET_FIXED_SIZE == 7, RPC_PACKET_PACKER.size


class RPCPacketType(Enum):
    Req = 0
    Rep = 1

    def __repr__(self):
        return f'{self.__class__.__name__}.{self.name}'


def json_pack(payload):
    try:
        return json.dumps(payload, separators=(',', ':')).encode()
    except:
        raise ValueError(f'Could not encode {payload}')


def json_unpack(payload: bytes):
    return json.loads(payload)


class RPCPacket(NamedTuple):
    key: RPCKey
    type: RPCPacketType
    name: str
    payload: Any = None
    """This must be json-serializable"""

    def pack(self) -> bytes:
        name = self.name.encode()
        payload = json_pack(self.payload)

        fixed_fields = RPC_PACKET_PACKER.pack(self.type.value, len(name), len(payload))
        body = self.key.pack() + fixed_fields + name + payload
        body = zlib.compress(body, level=9)
        return body

    @classmethod
    def unpack(cls, body: bytes):
        body = zlib.decompress(body)

        key = RPCKey.unpack(body[:RPC_KEY_SIZE])

        type_val, name_size, payload_size = RPC_PACKET_PACKER.unpack(body[RPC_KEY_SIZE:RPC_PACKET_HEADER_SIZE])
        type = RPCPacketType(type_val)
        name = body[RPC_PACKET_HEADER_SIZE:RPC_PACKET_HEADER_SIZE + name_size].decode()
        payload = body[RPC_PACKET_HEADER_SIZE + name_size:RPC_PACKET_HEADER_SIZE + name_size + payload_size]

        payload = json_unpack(payload)

        return RPCPacket(
            key,
            type,
            name,
            payload
        )



from dataclasses import dataclass
from typing import TypeVar, Generic, Type

from xrpc.dsl import rpc, regular
from xrpc.error import TerminationException
from xrpc.util import time_now

T = TypeVar('T')


@dataclass
class Data(Generic[T]):
    val: T


class GenericRPC(Generic[T]):
    def __init__(self, cls: Type[T], timeout=20.):
        self.cls = cls
        self.has_replied_a = False
        self.has_replied_b = False
        self.started = time_now()
        self.timeout = timeout

    @rpc()
    def process(self, val: T) -> Data[T]:
        self.has_replied_a = True
        return Data(val)

    @rpc()
    def process_blunt(self, val: T) -> T:
        self.has_replied_b = True
        return val

    @regular(tick=True)
    def tick(self) -> float:
        if (time_now() - self.started).total_seconds() > self.timeout:
            raise TimeoutError('Timeout')

        if all([self.has_replied_b, self.has_replied_a]):
            raise TerminationException()

        return 0.5



import datetime
import logging
from typing import NamedTuple, Any, Optional

from xrpc.const import SERVER_SERDE_INST
from xrpc.generic import build_generic_context
from xrpc.transform import RPCS, get_rpc
from xrpc.serde.abstract import SerdeStruct, SerdeSet
from xrpc.serde.types import CallableArgsWrapper, CallableRetWrapper


class ServiceDefn(NamedTuple):
    serde: SerdeStruct
    rpcs: RPCS
    rpcs_serde: Any

    @classmethod
    def from_obj(cls, obj):

        obj, ctx = build_generic_context(obj)

        rpcs = get_rpc(obj)
        rpcs_serde = {}

        serde_set: SerdeSet = SerdeSet.walk(SERVER_SERDE_INST, datetime.datetime, ctx)
        serde_set = serde_set.merge(SerdeSet.walk(SERVER_SERDE_INST, Optional[str], ctx))

        logging.getLogger(__name__).debug('Inferred RPCs: %s', rpcs)

        for rpc_name, rpc_def in rpcs.items():
            fa = CallableArgsWrapper.from_func_cls(obj, rpc_def.fn, )
            fb = CallableRetWrapper.from_func_cls(obj, rpc_def.fn, )

            new_serde_set1 = SerdeSet.walk(
                SERVER_SERDE_INST,
                fa,
                ctx,
            )

            new_serde_set2 = SerdeSet.walk(
                SERVER_SERDE_INST,
                fb,
                ctx,
            )

            new_serde_set = new_serde_set1.merge(new_serde_set2)

            new_serde_set = serde_set.merge(new_serde_set)

            serde_set = new_serde_set

            rpcs_serde[rpc_name] = [fa, fb]

        serde_struct: SerdeStruct = serde_set.struct(SERVER_SERDE_INST)

        return ServiceDefn(serde_struct, rpcs, rpcs_serde)



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, TypeVar, Generic

from xrpc.error import HorizonPassedError
from xrpc.net import RPCKey
from xrpc.util import time_now


class RPCLogDict:
    def __init__(self, horizon: datetime):
        self.horizon = horizon
        self.values: Dict[RPCKey, Any] = {}

    def __getitem__(self, item: RPCKey):
        if item.timestamp < self.horizon:
            raise HorizonPassedError(self.horizon)
        return self.values[item]

    def __contains__(self, item):
        try:
            _ = self[item]
            return True
        except KeyError:
            return False

    def __setitem__(self, key: RPCKey, value: Any):
        if key.timestamp < self.horizon:
            raise HorizonPassedError(self.horizon)

        self.values[key] = value

    def set_horizon(self, new_horizon: datetime):
        if new_horizon < self.horizon:
            raise HorizonPassedError(self.horizon)

        for key in [k for k in self.values.keys() if k.timestamp < new_horizon]:
            del self.values[key]

        self.horizon = new_horizon


K = TypeVar('K')
V = TypeVar('V')


@dataclass
class CachedDictRecord:
    ts: datetime
    v: V


class Hidrate(Generic[K, V]):
    def __call__(self, cd: 'CachedDict', k: K) -> V:
        raise NotImplementedError()


class CachedDict(Generic[K, V]):
    def __init__(self, hidrate_fn: Hidrate, max_timeout: timedelta = timedelta(seconds=120)):
        self.values: Dict[K, CachedDictRecord] = {}
        self.hidrate_fn: Hidrate = hidrate_fn
        self.max_timeout = max_timeout

    def __setitem__(self, key, value):
        self.values[key] = CachedDictRecord(time_now(), value)

    def __getitem__(self, item):
        now = time_now()

        if item in self.values:
            item_rec = self.values[item]

            if item_rec.ts > now - self.max_timeout:
                return item_rec.v
            else:
                return self.hidrate_fn(self, item)

        return self.hidrate_fn(self, item)

class ObjectDict:
    def __init__(self, **kwargs):
        self._dict = dict(**kwargs)

    def __setattr__(self, key, value):
        if key == '_dict':
            super().__setattr__(key, value)
        else:
            self._dict[key] = value

    def __getattr__(self, item):
        try:
            return self._dict[item]
        except KeyError:
            raise AttributeError()
